/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_base.h"
#include "types.h"
#include <stdio.h>

/**
 * h_relation_base_get_type:
 *
 * Returns the Glib type ID of #HRelationBase
 *
 * Return value: type ID
 */
G_DEFINE_TYPE (HRelationBase, h_relation_base, H_TYPE_RELATION)

static void h_relation_base_set_property (GObject      *object,
                                          guint         property,
                                          const GValue *value,
                                          GParamSpec   *spec);

static void h_relation_base_get_property (GObject      *object,
                                          guint         property,
                                          GValue       *value,
                                          GParamSpec   *spec);

static void h_relation_base_dispose      (GObject      *object);

static void h_relation_base_finalize     (GObject      *object);

static void h_relation_base_print        (HRelation *relation);

static void h_relation_base_for_each_attr (HRelation *relation,
                                           HAttrFunc  func,
                                           gpointer   user_data);

static void h_relation_base_for_each_tuple (HRelation  *relation,
                                            HTupleFunc  func,
                                            gpointer    user_data);

static guint h_relation_base_get_size    (HRelation *relation);

static guint h_relation_base_get_n_attrs (HRelation *relation);

static void
h_relation_base_class_init (HRelationBaseClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  object_class->set_property = h_relation_base_set_property;
  object_class->get_property = h_relation_base_get_property;
  object_class->dispose = h_relation_base_dispose;
  object_class->finalize = h_relation_base_finalize;

  relation_class->print = h_relation_base_print;
  relation_class->for_each_attr = h_relation_base_for_each_attr;
  relation_class->for_each_tuple = h_relation_base_for_each_tuple;
  relation_class->get_size = h_relation_base_get_size;
  relation_class->get_n_attrs = h_relation_base_get_n_attrs;
}

static void
h_relation_base_init (HRelationBase *self)
{
  /* nothing to do */
}

static void
h_relation_base_dispose (GObject *object)
{
  HRelationBase *self = H_RELATION_BASE (object);
  HRelation     *relation = H_RELATION (object);

  if (relation->dispose_has_run)
    return;

  relation->dispose_has_run = TRUE;

  g_free (self->primary_key);

  if (self->attrs)
    g_hash_table_destroy (self->attrs);

  if (self->tuples)
    g_hash_table_destroy (self->tuples);

  G_OBJECT_CLASS (h_relation_base_parent_class)->dispose (object);
}

static void
h_relation_base_finalize (GObject *object)
{
  G_OBJECT_CLASS (h_relation_base_parent_class)->finalize (object);
}

static void
h_relation_base_set_property (GObject      *object,
                              guint         property,
                              const GValue *value,
                              GParamSpec   *spec)
{}

static void
h_relation_base_get_property (GObject      *object,
                              guint         property,
                              GValue       *value,
                              GParamSpec   *spec)
{}

static HAttribute *
attr_new (const gchar *name,
          GType        type)
{
  HAttribute *attr;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (type != G_TYPE_NONE, NULL);

  attr = g_new (HAttribute, 1);

  if (attr)
    {
      attr->name = g_strdup (name);
      attr->type = type;
    }

  return attr;
}

static void
attr_free (HAttribute *attr)
{
  g_free (attr->name);
  g_free (attr);
}

static HAttribute *
attr_map_get (GHashTable  *map,
              const gchar *name)
{
  g_return_val_if_fail (map != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return (HAttribute *) g_hash_table_lookup (map, name);
}

#if 0
static HAttribute *
relation_get_attr (HRelation   *relation,
                   const gchar *name)
{
  g_return_val_if_fail (relation != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return attr_map_get (relation->attrs, name);
}
#endif

static GHashTable *
build_attr_map (va_list *args)
{
  GHashTable  *map;
  const gchar *name;
  GType        type;
  HAttribute  *attr;

  /* Collect attribute list into a name-attribute associative map */

  map = g_hash_table_new_full (g_str_hash, g_str_equal, NULL,
                               (GDestroyNotify) attr_free);
  if (!map)
    return NULL;

  while (1)
    {
      name = va_arg (*args, const gchar *);
      if (!name)
        break;

      type = va_arg (*args, GType);

      if (g_hash_table_lookup (map, name) || !h_type_known (type))
        {
          g_hash_table_destroy (map);
          return NULL;
        }

      attr = attr_new (name, type);
      g_hash_table_insert (map, attr->name, attr);
    }

  return map;
}

static const gchar *
build_key (va_list *args)
{
  return va_arg (*args, const gchar *);
}

/**
 * h_relation_base_new:
 * @name: name
 * @description: description
 * @...: Attribute name followed by its type, then another attribute
 * name followed by its type, and so on, then NULL, followed by a
 * primary key attribute name, followed by another, then NULL.
 *
 * Returns a new base relation (#HRelationBase) with name @name,
 * description @description, specified attributes and primary key
 * attributes.
 *
 * Return value: a #HRelationBase or %NULL
 **/
HRelation *
h_relation_base_new (const gchar *name,
                     const gchar *description,
                     ...)
{
  HRelationBase *relation;
  GHashTable   *attrs, *tuples;
  const gchar  *primary_key;
  GType         key_type;
  va_list       args;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (description != NULL, NULL);

  va_start (args, description);

  attrs = build_attr_map (&args);
  primary_key = build_key (&args);

  va_end (args);

  /* make sure we have at least 1 attribute and that primary key
     attribute is valid */
  if (g_hash_table_size (attrs) == 0 ||
      !attr_map_get (attrs, primary_key))
    {
      g_hash_table_destroy (attrs);
      return NULL;
    }

  key_type = attr_map_get (attrs, primary_key)->type;

  tuples = g_hash_table_new_full (h_type_hash_func (key_type),
                                  h_type_equal_func (key_type),
                                  NULL,
                                  (GDestroyNotify) h_tuple_free);

  relation = g_object_new (H_TYPE_RELATION_BASE, NULL);

  g_object_set (relation,
                "name", name,
                "description", description,
                NULL);

  relation->primary_key = g_strdup (primary_key);
  relation->attrs = attrs;
  relation->tuples = tuples;

  return H_RELATION (relation);
}

typedef struct
{
  HTuple   *tuple;
  gboolean  match;
}
CheckTupleContext;

static void
check_tuple (const gchar       *key,
             const HAttribute  *attr,
             CheckTupleContext *context)
{
  HTupleElem *elem;

  if (!context->match)
    return;

  elem = h_tuple_get (context->tuple, key);

  if (!elem)
    {
      context->match = FALSE;
      return;
    }

  if (elem->type != G_TYPE_NONE && elem->type != attr->type)
    context->match = FALSE;
}

static gboolean
relation_check_tuple (HRelationBase *self,
                      HTuple        *tuple)
{
  CheckTupleContext context;

  if (h_tuple_get_size (tuple) !=
      h_relation_base_get_n_attrs ((HRelation *) self))
    {
      h_tuple_free (tuple);
      return FALSE;
    }

  context.tuple = tuple;
  context.match = TRUE;

  g_hash_table_foreach (self->attrs, (GHFunc) check_tuple, &context);

  return context.match;
}

typedef struct
{
  HRelationBase *relation;
  HTuple        *tuple;
  HTuple        *copy;
}
CopyTupleContext;

static void
copy_tuple (const gchar      *key,
            const HAttribute *attr,
            CopyTupleContext *context)
{
  h_tuple_insert (context->copy, key, attr->type,
                h_tuple_get (context->tuple, key)->value);
}

static void
insert_tuple (HRelationBase *self,
              HTuple        *tuple)
{
  CopyTupleContext context;

  context.relation = self;
  context.tuple = tuple;
  context.copy = h_tuple_new ();

  g_hash_table_foreach (self->attrs, (GHFunc) copy_tuple, &context);

  g_hash_table_insert (self->tuples,
                       h_tuple_get (context.copy, self->primary_key)->value,
                       context.copy);

  g_signal_emit_by_name (self, "insert-tuple", context.copy);
}

/**
 * h_relation_base_insert_tuple:
 * @relation: a #HRelationBase
 * @...: An attribute name, followed by its coreesponding value, followed by
 * another attribute name, and so on.
 *
 * Inserts a tuple into @relation.
 *
 * Return value: TRUE if tuple was successfully inserted, FALSE otherwise
 **/
gboolean
h_relation_base_insert_tuple (HRelationBase *self,
                              ...)
{
  va_list  args;
  HTuple  *tuple;
  gint     n_attrs;

  g_return_val_if_fail (self != NULL, FALSE);

  n_attrs = h_relation_base_get_n_attrs ((HRelation *) self);

  /* collect tuples into a tuple */
  va_start (args, self);
  tuple = h_tuple_new_v (n_attrs, args);
  va_end (args);

  /* type-check tuple against relation */
  if (!relation_check_tuple (self, tuple))
    {
      h_tuple_free (tuple);
      return FALSE;
    }

  insert_tuple (self, tuple);
  h_tuple_free (tuple);

  return TRUE;
}

/**
 * h_relation_base_get_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Returns a tuple in @relation that has the primary key @key.
 *
 * Return value: a #HTuple storing the tuple or %NULL if no tuples have
 * the specified key.
 */
HTuple *
h_relation_base_get_tuple (HRelationBase *self,
                           gconstpointer  key)
{
  HTuple *tuple;

  g_return_val_if_fail (H_IS_RELATION_BASE (self), FALSE);

  tuple = (HTuple *) g_hash_table_lookup (self->tuples, key);

  return tuple ? h_tuple_clone (tuple) : NULL;
}

/**
 * h_relation_base_delete_tuple:
 * @relation: a #HRelationBase
 * @key: primary key value
 *
 * Deletes a tuple in @relation that has the primary key @key.
 *
 * Return value: TRUE if tuple was deleted, FALSE if no such tuple
 * exists.
 **/
gboolean
h_relation_base_delete_tuple (HRelationBase *self,
                              gconstpointer  key)
{
  HTuple *tuple;

  g_return_val_if_fail (self != NULL, FALSE);

  tuple = g_hash_table_lookup (self->tuples, key);
  g_return_val_if_fail (tuple != NULL, FALSE);

  g_signal_emit_by_name (self, "delete-tuple", tuple);

  return g_hash_table_remove (self->tuples, key);
}

/**
 * h_relation_base_get_primary_key:
 * @relation: a #HRelationBase
 *
 * Returns the primary key attributes of @relation.
 *
 * Return value: primary key attribute
 **/
const gchar *
h_relation_base_get_primary_key (HRelationBase *self)
{
  return self->primary_key;
}

static void
h_relation_base_print (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_if_fail (self != NULL);

  printf ("Primary key: %s\n", self->primary_key);
}

typedef struct
{
  HAttrFunc func;
  gpointer  data;
}
IterAttrContext;

static void
iter_attr (gconstpointer    key,
           HAttribute      *attr,
           IterAttrContext *context)
{
  context->func (attr, context->data);
}

static void
h_relation_base_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationBase *self = H_RELATION_BASE (relation);
  IterAttrContext context;

  g_return_if_fail (self != NULL);

  context.func = func;
  context.data = user_data;
  g_hash_table_foreach (self->attrs, (GHFunc) iter_attr, &context);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
IterTupleContext;

static void
iter_tuple (gconstpointer     key,
            HTuple           *tuple,
            IterTupleContext *context)
{
  context->func (tuple, context->data);
}

static void
h_relation_base_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationBase *self = H_RELATION_BASE (relation);
  IterTupleContext context;

  g_return_if_fail (self != NULL);

  context.func = func;
  context.data = user_data;
  g_hash_table_foreach (self->tuples, (GHFunc) iter_tuple, &context);
}

static guint
h_relation_base_get_size (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_val_if_fail (self != NULL, 0);

  return g_hash_table_size (self->tuples);
}

static guint
h_relation_base_get_n_attrs (HRelation *relation)
{
  HRelationBase *self = H_RELATION_BASE (relation);

  g_return_val_if_fail (self != NULL, 0);

  return g_hash_table_size (self->attrs);
}
