/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_restrict.h"
#include <stdio.h>

/**
 * TuplePredFunc:
 * @tuple: a #HTuple
 * @user_data: user data
 *
 * A #HTuple predicate function.
 *
 * Return value: either TRUE or FALSE
 **/

/**
 * h_relation_restrict_get_type:
 *
 * Returns the Glib type ID of HRelationRestrict
 *
 * Return value: type ID
 **/
G_DEFINE_TYPE (HRelationRestrict, h_relation_restrict, H_TYPE_RELATION)

static void h_relation_restrict_dispose  (GObject *object);

static void h_relation_restrict_finalize (GObject *object);

static void h_relation_restrict_print (HRelation *relation);

static void h_relation_restrict_for_each_attr (HRelation *relation,
                                               HAttrFunc  func,
                                               gpointer   user_data);

static void h_relation_restrict_for_each_tuple (HRelation  *relation,
                                                HTupleFunc  func,
                                                gpointer    user_data);

static guint h_relation_restrict_get_size (HRelation *relation);

static guint h_relation_restrict_get_n_attrs (HRelation *relation);

static void h_relation_restrict_insert_tuple (HRelationSort *self,
                                              HTuple        *tuple);

static void h_relation_restrict_delete_tuple (HRelationSort *self,
                                              HTuple        *tuple);

static void
h_relation_restrict_class_init (HRelationRestrictClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *h_relation_class = H_RELATION_CLASS (klass);

  object_class->dispose  = h_relation_restrict_dispose;
  object_class->finalize = h_relation_restrict_finalize;

  h_relation_class->print = h_relation_restrict_print;
  h_relation_class->for_each_attr = h_relation_restrict_for_each_attr;
  h_relation_class->for_each_tuple = h_relation_restrict_for_each_tuple;
  h_relation_class->get_size = h_relation_restrict_get_size;
  h_relation_class->get_n_attrs = h_relation_restrict_get_n_attrs;
}

static void
h_relation_restrict_init (HRelationRestrict *self)
{
  self->to_restrict = NULL;
  self->func = NULL;
  self->func_data = NULL;
}

static void
h_relation_restrict_dispose (GObject *object)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (object);
  HRelation *relation = H_RELATION (object);

  if (relation->dispose_has_run)
    return;

  relation->dispose_has_run = TRUE;

  g_signal_handlers_disconnect_by_func (self->to_restrict,
                                        (gpointer) h_relation_restrict_insert_tuple,
                                        self);

  g_signal_handlers_disconnect_by_func (self->to_restrict,
                                        (gpointer) h_relation_restrict_delete_tuple,
                                        self);

  g_hash_table_destroy (self->tuples);

  g_object_unref (self->to_restrict);
}

static void
h_relation_restrict_finalize (GObject *object)
{
  /* do nothing */
}

static void
h_relation_restrict_print (HRelation *relation)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);
  gchar *str;

  g_return_if_fail (self != NULL);

  printf ("Relation to restrict:\n");

  str = h_relation_get_name (self->to_restrict);
  printf ("  Name: %s\n", str ? str : "(none)");
  g_free (str);

  str = h_relation_get_desc (self->to_restrict);
  printf ("  Description: %s\n", str ? str : "(none)");
  g_free (str);
}

static guint
h_relation_restrict_get_n_attrs (HRelation *relation)
{
  HRelationRestrict *self = H_RELATION_RESTRICT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_n_attrs (self->to_restrict);
}

static guint
h_relation_restrict_get_size (HRelation *relation)
{
  /* do nothing */
  return 0;
}

static void
h_relation_restrict_for_each_attr (HRelation *relation,
                                   HAttrFunc  func,
                                   gpointer   user_data)
{
  g_return_if_fail (H_IS_RELATION_RESTRICT (relation));
}

static void
h_relation_restrict_for_each_tuple (HRelation  *relation,
                                    HTupleFunc  func,
                                    gpointer    user_data)
{
  g_return_if_fail (H_IS_RELATION_RESTRICT (relation));
}

static void
h_relation_restrict_insert_tuple (HRelationSort *self,
                                  HTuple        *tuple)
{

  g_signal_emit_by_name (self, "insert-tuple", tuple);
}

static void
h_relation_restrict_delete_tuple (HRelationSort *self,
                                  HTuple        *tuple)
{
  g_signal_emit_by_name (self, "delete-tuple", tuple);
}

static void
insert_tuple (HTuple            *tuple,
              HRelationRestrict *self)
{
}

static void
init_tuples (HRelationRestrict *self)
{
  self->tuples = g_hash_table_new (g_direct_hash, (GEqualFunc) h_tuple_equal);

  h_relation_for_each_tuple (self->to_restrict, (HTupleFunc) insert_tuple, self);
}

/**
 * h_relation_restrict:
 * @relation: a #HRelation to perform restrict on
 * @predicate_func: tuple predicate function
 * @user_data: Custom data to pass to @predicate_func
 *
 * Creates a 'restricted' relation, containing every tuple in
 * @relation that evaluates to TRUE with @predicate_func.
 *
 * Return value: a #HRelationRestrict
 **/
HRelation *
h_relation_restrict (HRelation     *to_restrict,
                     HTuplePredFunc predicate_func,
                     gpointer       user_data)
{
  HRelationRestrict *relation;

  g_return_val_if_fail (to_restrict != NULL, NULL);
  g_return_val_if_fail (predicate_func != NULL, NULL);
  g_return_val_if_fail (user_data != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_RESTRICT, NULL);

  relation->to_restrict = to_restrict;
  g_object_ref (to_restrict);

  relation->func = predicate_func;
  relation->func_data = user_data;

  init_tuples (relation);

  g_signal_connect_swapped (to_restrict, "insert-tuple",
                            G_CALLBACK (h_relation_restrict_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_restrict, "delete-tuple",
                            G_CALLBACK (h_relation_restrict_delete_tuple),
                            relation);

  return NULL;
}
