/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "relation_sort.h"
#include <stdio.h>

/**
 * h_relation_sort_get_type:
 *
 * Returns the Glib type ID of #HRelationSort.
 *
 * Return value: type ID
 **/
G_DEFINE_TYPE (HRelationSort, h_relation_sort, H_TYPE_RELATION)

static void h_relation_sort_dispose  (GObject *object);

static void h_relation_sort_finalize (GObject *object);

static void h_relation_sort_for_each_attr (HRelation *relation,
                                           HAttrFunc  func,
                                           gpointer   user_data);

static void h_relation_sort_for_each_tuple (HRelation  *relation,
                                            HTupleFunc  func,
                                            gpointer    user_data);

static void h_relation_sort_print (HRelation *relation);

static guint h_relation_sort_get_size    (HRelation *relation);

static guint h_relation_sort_get_n_attrs (HRelation *relation);

static void h_relation_sort_insert_tuple (HRelationSort *self,
                                          HTuple        *tuple);

static void h_relation_sort_delete_tuple (HRelationSort *self,
                                          HTuple        *tuple);

static void
h_relation_sort_class_init (HRelationSortClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  HRelationClass *relation_class = H_RELATION_CLASS (klass);

  object_class->dispose  = h_relation_sort_dispose;
  object_class->finalize = h_relation_sort_finalize;

  relation_class->print = h_relation_sort_print;
  relation_class->for_each_attr = h_relation_sort_for_each_attr;
  relation_class->for_each_tuple = h_relation_sort_for_each_tuple;
  relation_class->get_size = h_relation_sort_get_size;
  relation_class->get_n_attrs = h_relation_sort_get_n_attrs;
}

static void
h_relation_sort_init (HRelationSort *self)
{
  self->to_sort = NULL;
  self->compare_func = NULL;
  self->order = NULL;
}

static void
h_relation_sort_dispose (GObject *object)
{
  HRelationSort *self = H_RELATION_SORT (object);
  HRelation     *relation = H_RELATION (object);

  if (relation->dispose_has_run)
    return;

  relation->dispose_has_run = TRUE;

  g_signal_handlers_disconnect_by_func (self->to_sort,
                                        (gpointer) h_relation_sort_insert_tuple,
                                        self);


  g_signal_handlers_disconnect_by_func (self->to_sort,
                                        (gpointer) h_relation_sort_delete_tuple,
                                        self);

  g_object_unref (self->to_sort);

  g_tree_destroy (self->order);
}

static void
h_relation_sort_finalize (GObject *object)
{
  /* do nothing */
}

static void
h_relation_sort_print (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);
  gchar *str;

  g_return_if_fail (self != NULL);

  printf ("HRelation to sort:\n");

  str = h_relation_get_name (self->to_sort);
  printf ("  Name: %s\n", str ? str : "(none)");
  g_free (str);

  str = h_relation_get_desc (self->to_sort);
  printf ("  Description: %s\n", str ? str : "(none)");
  g_free (str);
}

static void
h_relation_sort_for_each_attr (HRelation *relation,
                               HAttrFunc  func,
                               gpointer   user_data)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_if_fail (self != NULL);

  h_relation_for_each_attr (self->to_sort, func, user_data);
}

typedef struct
{
  HTupleFunc func;
  gpointer   data;
}
OrderMapIterateContext;

static void
order_map_iterate (gpointer                key,
                   GList                  *list,
                   OrderMapIterateContext *context)
{
  GList *iter;

  for (iter = list; iter != NULL; iter = g_list_next (iter))
    {
      context->func ((HTuple *) iter->data, context->data);
    }
}

static void
h_relation_sort_for_each_tuple (HRelation  *relation,
                                HTupleFunc  func,
                                gpointer    user_data)
{
  HRelationSort *self = H_RELATION_SORT (relation);
  OrderMapIterateContext context;

  g_return_if_fail (self != NULL);

  context.func = func;
  context.data = user_data;

  g_tree_foreach (self->order, (GTraverseFunc) order_map_iterate, &context);
}


static guint
h_relation_sort_get_size (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_size (self->to_sort);
}

static guint
h_relation_sort_get_n_attrs (HRelation *relation)
{
  HRelationSort *self = H_RELATION_SORT (relation);

  g_return_val_if_fail (self != NULL, 0);

  return h_relation_get_n_attrs (self->to_sort);
}

static void
order_map_insert (HTuple        *tuple,
                  HRelationSort *relation)
{
  GList *list;

  /* FIXME: should store the key instead using the tuple pointer
     directly */

  list = g_tree_lookup (relation->order, tuple);

  if (list)
    g_tree_steal (relation->order, tuple);

  list = g_list_append (list, tuple);
  g_tree_insert (relation->order, tuple, list);
}


static GList *
list_find_custom_with_data (GList            *list,
                            gconstpointer     data,
                            GCompareDataFunc  func,
                            gpointer          func_data)
{
  GList *iter;

  for (iter = list; iter != NULL; iter = g_list_next (iter))
    {
      if (func (data, iter->data, func_data) == 0)
        return iter;
    }

  return NULL;
}

static void
order_map_delete (HTuple        *tuple,
                  HRelationSort *relation)
{
  GList *list, *node;

  list = g_tree_lookup (relation->order, tuple);
  if (!list)
    return;

  node = list_find_custom_with_data (list, tuple,
                                     relation->compare_func,
                                     relation->compare_func_data);
  if (!node)
    return;

  g_list_remove_link (list, node);
  g_list_free_1 (node);
}

static void
h_relation_sort_insert_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  order_map_insert (tuple, self);

  g_signal_emit_by_name (self, "insert-tuple", tuple);
}

static void
h_relation_sort_delete_tuple (HRelationSort *self,
                              HTuple        *tuple)
{
  order_map_delete (tuple, self);

  g_signal_emit_by_name (self, "delete-tuple", tuple);
}

static void
build_order_map (HRelationSort *self)
{
  self->order = g_tree_new_full (self->compare_func,
                                 self->compare_func_data,
                                 NULL,
                                 (GDestroyNotify) g_list_free);

  h_relation_for_each_tuple (self->to_sort,
                             (HTupleFunc) order_map_insert,
                             self);
}

/**
 * h_relation_sort:
 * @relation: a #HRelation to sort
 * @compare_func: function to compare a pair of tuples in @relation
 * @user_data: Custom data passed to @compare_func
 *
 * Creates a sorted #HRelation.
 *
 * Return value: a #HRelationSort
 **/
HRelation *
h_relation_sort (HRelation        *to_sort,
                 GCompareDataFunc  compare_func,
                 gpointer          user_data)
{
  HRelationSort *relation;

  g_return_val_if_fail (to_sort != NULL, NULL);
  g_return_val_if_fail (compare_func != NULL, NULL);

  relation = g_object_new (H_TYPE_RELATION_SORT, NULL);

  relation->to_sort = to_sort;
  g_object_ref (to_sort);

  relation->compare_func = compare_func;
  relation->compare_func_data = user_data;

  g_signal_connect_swapped (to_sort, "insert-tuple",
                            G_CALLBACK (h_relation_sort_insert_tuple),
                            relation);

  g_signal_connect_swapped (to_sort, "delete-tuple",
                            G_CALLBACK (h_relation_sort_delete_tuple),
                            relation);

  build_order_map (relation);

  return H_RELATION (relation);
}

/**
 * h_relation_sort_by_attrs:
 * @relation: a #HRelation to sort
 * @...: List of names of attributes to compare in order, ending with
 * %NULL.
 *
 * Creates a #HRelation sorted by attributes.
 *
 * Return value: a #HRelationSort
 **/
HRelation *
h_relation_sort_by_attrs (HRelation *relation,
                          ...)
{
  /* FIXME: implement */

  return NULL;
}
