/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "tuple.h"
#include "types.h"
#include <stdio.h>

/**
 * HTupleElem:
 * @type: type
 * @value: value of @type
 *
 * Stores type and value of a tuple (#HTuple) component.
 **/

static void print_elem (const gchar *key,
                        HTupleElem  *elem);

static void copy_elem  (const gchar *name,
                        HTupleElem  *elem,
                        HTuple      *tuple);

static void free_elem  (HTupleElem  *elem);

static gpointer copy_value (GType         type,
                            gconstpointer value);

/**
 * h_tuple_get_type:
 *
 * Returns the Glib type ID for #HTuple.
 *
 * Return value: type ID for #HTuple
 **/
GType
h_tuple_get_type (void)
{
  static GType type = G_TYPE_INVALID;

  if (!type)
    {
      type = g_boxed_type_register_static ("HTuple",
                                           (GBoxedCopyFunc) h_tuple_clone,
                                           (GBoxedFreeFunc) h_tuple_free);
    }

  return type;
}

/**
 * h_tuple_new:
 *
 * Creates a new empty tuple.
 *
 * Return value: a newly created empty #HTuple, or %NULL.
 **/
HTuple *
h_tuple_new (void)
{
  return g_hash_table_new_full (g_str_hash, g_str_equal,
                                (GDestroyNotify) g_free,
                                (GDestroyNotify) free_elem);
}

/**
 * h_tuple_insert:
 * @tuple: a #HTuple
 * @name: name of attribute
 * @type: type of attribute
 * @value: value of @type
 *
 * Adds a new atribute named @name, of type @type and value @value.
 **/
void
h_tuple_insert (HTuple        *tuple,
                const gchar  *name,
                GType         type,
                gpointer      value)
{
  HTupleElem *elem;

  g_return_if_fail (tuple != NULL);
  g_return_if_fail (name != NULL);

  if (type != G_TYPE_NONE)
    g_return_if_fail (h_type_known (type));

  elem = g_new (HTupleElem, 1);
  g_return_if_fail (elem != NULL);

  elem->type = type;
  elem->value = copy_value (type, value);

  g_hash_table_insert (tuple, g_strdup (name), elem);
}

/**
 * h_tuple_new_v:
 * @size: degree
 * @list: argument list
 *
 * Constructs a tuple from an argument list.
 *
 * Return value: a newly constructed #HTuple, or %NULL.
 **/
HTuple *
h_tuple_new_v (guint   size,
               va_list list)
{
  HTuple         *tuple;
  gchar         *name;
  gconstpointer *value;
  int            i;

  tuple = h_tuple_new ();
  if (!tuple)
    return NULL;

  for (i = 0; i < size; i++)
    {
      name  = va_arg (list, gchar *);
      value = va_arg (list, gpointer);

      h_tuple_insert (tuple, name, G_TYPE_NONE, value);
    }

  return tuple;
}

/**
 * h_tuple_get:
 * @tuple: a #HTuple
 * @name: name of attribute
 *
 * Returns the type and value of an attribute in @tuple, or %NULL if
 * attribute doesn't exist.
 *
 * Return value: a #HTupleElem containing the type and value.
 **/
HTupleElem *
h_tuple_get (HTuple       *tuple,
             const gchar *name)
{
  g_return_val_if_fail (tuple != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  return (HTupleElem *) g_hash_table_lookup (tuple, name);
}

/**
 * h_tuple_get_value:
 * @tuple: a #HTuple
 * @name: name of attribute
 *
 * Returns the value of an attribute in @tuple, or %NULL if attribute
 * doesn't exist.
 *
 * Return value: a copy of the attribute value
 **/
gpointer
h_tuple_get_value (HTuple       *tuple,
                   const gchar *name)
{
  HTupleElem *elem;

  g_return_val_if_fail (tuple != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);

  elem = h_tuple_get (tuple, name);
  return elem ? h_value_clone (elem->type, elem->value) : NULL;
}

/**
 * h_tuple_get_size:
 * @tuple: a #HTuple
 *
 * Returns the degree ('size') of @tuple.
 *
 * Return value: degree ('size') of @tuple
 **/
guint
h_tuple_get_size (HTuple *tuple)
{
  g_return_val_if_fail (tuple != NULL, 0);

  return g_hash_table_size (tuple);
}

/**
 * h_tuple_clone:
 * @tuple: a #HTuple
 *
 * Returns a copy of @tuple.
 *
 * Return value: copy of @tuple
 **/
HTuple *
h_tuple_clone (HTuple *tuple)
{
  HTuple *copy;

  g_return_val_if_fail (tuple != NULL, NULL);

  copy = h_tuple_new ();
  g_hash_table_foreach (tuple, (GHFunc) copy_elem, copy);
  return copy;
}

/**
 * h_tuple_free:
 * @tuple: a #HTuple
 *
 * Free memory used by @tuple.
 **/
void
h_tuple_free (HTuple *tuple)
{
  if (tuple)
    g_hash_table_destroy (tuple);
}

/**
 * h_tuple_print:
 * @tuple: a #HTuple
 *
 * Prints the content of @tuple to stdout.
 **/
void
h_tuple_print (HTuple *tuple)
{
  g_return_if_fail (tuple != NULL);
  printf ("(");
  g_hash_table_foreach (tuple, (GHFunc) print_elem, NULL);
  printf (")");
}

/**
 * direct_compare:
 * @a: left operand
 * @b: right operand
 *
 * Compares @a and @b using the builtin C operators <, and >.
 *
 * Return value: -1 if @a is greater, 1 if @b is greater, 0 if equal
 **/
gint
direct_compare (gconstpointer a,
                gconstpointer b)
{
  if ((guint) a > (guint) b)
    return 1;

  if ((guint) a < (guint) b)
    return -1;

  return 0;
}

/**
 * h_tuple_compare_elem:
 * @tuple1: left #HTuple operand
 * @tuple2: right #HTuple operand
 * @name: name of attribute to compare
 *
 * Compares the values of a common attribute in @tuple1 and @tuple2.
 *
 * Return value: -1 if @tuple1 is greater, 1 if @tuple2 is greater, 0 if equal
 **/
gint
h_tuple_compare_elem (HTuple      *tuple1,
                      HTuple      *tuple2,
                      const gchar *name)
{
  /* NOTE: this function does not type-check the tuple operands */

  GType type;
  gpointer value1, value2;

  g_return_val_if_fail (tuple1 != NULL, 0);
  g_return_val_if_fail (tuple2 != NULL, 0);

  type = h_tuple_get (tuple1, name)->type;

  value1 = h_tuple_get (tuple1, name)->value;
  value2 = h_tuple_get (tuple2, name)->value;

  return (type != G_TYPE_NONE) ?
    h_value_compare (type, value1, value2) :
    direct_compare (value1, value2);
}

/**
 * h_tuple_compare_elems_v:
 * @tuple1: left #HTuple operand
 * @tuple2: right #HTuple operand
 * @elem_names: argument list (see #h_tuple_compare_elems)
 *
 * Compares the values of a common attribute in @tuple1 and @tuple2.
 *
 * Return value: -1 if @tuple1 is greater, 1 if @tuple2 is greater, 0 if equal
 **/
gint
h_tuple_compare_elems_v (HTuple  *tuple1,
                         HTuple  *tuple2,
                         va_list  elem_names)
{
  /* NOTE: this function does not type-check the tuple operands */

  const gchar *elem_name;
  gint result;

  g_return_val_if_fail (tuple1 != NULL, 0);
  g_return_val_if_fail (tuple2 != NULL, 0);

  while (1)
    {
      elem_name = va_arg (elem_names, const gchar *);
      if (!elem_name)
        break;

      result = h_tuple_compare_elem (tuple1, tuple2, elem_name);

      if (result)
        return result;
    }

  return 0;
}

/**
 * h_tuple_compare_elems:
 * @tuple1: left #HTuple operand
 * @tuple2: right #HTuple operand
 * @...: name of 1st attribute, name of 2nd attribute, and so on, ending with a %NULL.
 *
 * Compares the values of each common attribute in @tuple1 and @tuple2 in
 * the order listed. (TODO: needs elaboration)
 *
 * Return value: -1 if @tuple1 is greater, 1 if @tuple2 is greater, 0 if equal
 **/
gint
h_tuple_compare_elems (HTuple *tuple1,
                       HTuple *tuple2,
                       ...)
{
  va_list args;
  gint result;

  va_start (args, tuple2);
  result = h_tuple_compare_elems_v (tuple1, tuple2, args);
  va_end (args);

  return result;
}

typedef struct
{
  HTuple    *tuple;
  gboolean  result;
}
HTupleElemCompareContext;

static void
h_tuple_elem_compare (const gchar              *elem_name,
                      const HTupleElem         *elem,
                      HTupleElemCompareContext *context)
{
  HTupleElem *other;

  if (!context->result)
    return;

  if (elem->type == G_TYPE_NONE)
    {
      context->result = FALSE;
      return;
    }

  other = h_tuple_get (context->tuple, elem_name);
  if (elem->type != other->type)
    {
      context->result = FALSE;
      return;
    }

  context->result = h_value_compare (elem->type, elem->value, other->value);
}

/**
 * h_tuple_equal:
 * @tuple1: a #HTuple
 * @tuple2: a #HTuple
 *
 * Compares two @tuple1 and @tuple2 for equality. (TODO: Need elaboration).
 *
 * Return value: TRUE if equal, FALSE if not
 **/
gboolean
h_tuple_equal (HTuple *tuple1,
               HTuple *tuple2)
{
  HTupleElemCompareContext context;

  context.tuple = tuple2;
  context.result = FALSE;
  g_hash_table_foreach (tuple1, (GHFunc) h_tuple_elem_compare, &context);

  return context.result;
}

static void
combine_hash (const gchar      *elem_name,
              const HTupleElem *elem,
              guint            *hash)
{
  *hash ^= (*h_type_hash_func (elem->type)) (elem->value);
}

/**
 * h_tuple_hash:
 * @tuple: a #HTuple
 *
 * Returns the hash of @tuple.
 *
 * Return value: hash number
 **/
guint
h_tuple_hash (HTuple *tuple)
{
  guint hash = 0;

  g_hash_table_foreach (tuple, (GHFunc) combine_hash, &hash);
  return hash;
}

/**
 * h_tuple_hash_elems:
 * @tuple: a #HTuple
 * @...: first attribute name, followed by the second, and so on, ending with a %NULL.
 *
 * Returns the hash of a subset of @tuple. Note that the hash is
 * independent of the order in which attribute names are
 * listed. (TODO: elaborate).
 *
 * Return value: hash number
 **/
guint
h_tuple_hash_elems (HTuple *tuple,
                    ...)
{
  va_list          args;
  const gchar     *elem_name;
  const HTupleElem *elem;
  guint            hash = 0;

  va_start (args, tuple);

  while (1)
    {
      elem_name = va_arg (args, const gchar *);
      if (!elem_name)
        break;

      elem = h_tuple_get (tuple, elem_name);

      hash ^= (*h_type_hash_func (elem->type)) (elem->value);
    }

  va_end (args);

  return hash;
}

static void
print_elem (const gchar *key,
            HTupleElem  *elem)
{
  if (elem->type != G_TYPE_NONE)
    {
      gchar *str = h_value_str (elem->type, elem->value);
      printf (" %s:%s:%s ", key, g_type_name (elem->type), str);
      g_free (str);
    }
  else
    printf (" %s::0x%d ", key, GPOINTER_TO_INT (elem->value));
}

static void
copy_elem (const gchar *name,
           HTupleElem  *elem,
           HTuple      *tuple)
{
  HTupleElem *copy = g_new (HTupleElem, 1);

  copy->type = elem->type;
  copy->value = copy_value (elem->type, elem->value);

  g_hash_table_insert (tuple, g_strdup (name), copy);
}

static void
free_elem (HTupleElem *elem)
{
  g_return_if_fail (elem != NULL);

  if (elem->type != G_TYPE_NONE)
    h_value_free (elem->type, elem->value);

  g_free (elem);
}

static gpointer
copy_value (GType         type,
            gconstpointer value)
{
  return (type != G_TYPE_NONE) ?
    h_value_clone (type, value) :
    (gpointer) value;
}
