#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <getopt.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <glib-object.h>
#include <glib/gstdio.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#include "ui_playlist.h"
#endif

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xmlreader.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>

#include "main.h"
#include "configuration.h"

#include "bmpx/xml.h"

#define ROOT_NODE_NAME "bmp"

G_DEFINE_TYPE (BmpConfiguration, bmp_configuration, G_TYPE_OBJECT)
typedef enum {
         BMP_CONFIGURATION_SIGNAL_CHANGED,
         BMP_CONFIGURATION_SIGNAL_REMOVED,
         BMP_CONFIGURATION_SIGNAL_ADDED,

         N_SIGNALS
} BmpConfigurationSignals;

static guint signals[N_SIGNALS] = { 0, 0, 0 };

typedef enum {
    N_PROPS
} BmpConfigurationProperties;

struct _BmpConfigurationPrivate {
    gboolean dispose_has_run;

    xmlDocPtr xmlconfig;
};

#ifdef HAVE_GUI

typedef struct {
    GObject *object;
    BindingDataFunc func;    
    gchar *domain;
    gchar *key;
} GObjectBinding;

static void
toggle_action_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_BOOLEAN);
  g_value_set_boolean (value, gtk_toggle_action_get_active (GTK_TOGGLE_ACTION(object)));
}

static void
toggle_button_binding (GObject *object, GValue *value)
{
  g_value_init (value, G_TYPE_BOOLEAN);
  g_value_set_boolean (value, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(object)));
}

static void
binding_object_signal_handler_invoked (GObject *object, gpointer user_data)
{
    GObjectBinding *binding;
    GValue value = {0,};

    g_return_if_fail (G_IS_OBJECT(object));

    binding = (GObjectBinding*)user_data; 

    binding->func (object, &value);

    bmp_configuration_key_set (bmp_configuration,
			       binding->domain,
			       binding->key,
			       &value,
			       TRUE);
}

static void
object_binding_marshal (GClosure     *closure,
			GValue       *return_value,
			guint         n_param_values,
			const GValue *param_values,
			gpointer      invocation_hint,
			gpointer      marshal_data)
{
  typedef gboolean (*GMarshalBindingFunc) (gpointer     data1,
					   gpointer     data2);

  register GMarshalBindingFunc callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  data1 = g_value_peek_pointer (param_values + 0);
  data2 = closure->data;

  callback = (GMarshalBindingFunc) (cc->callback);

  callback (data1,
	    data2);
}

void
bmp_configuration_bind_object (BmpConfiguration *self,
			       GObject		*object, 
			       BindingDataFunc	 func, 
			       const gchar	*signal_spec, 
			       const gchar	*domain, 
			       const gchar	*key)
{
    GClosure *closure;
    GObjectBinding *binding;

    g_return_if_fail (BMP_IS_CONFIGURATION(self));

    binding = g_new0 (GObjectBinding, 1);

    binding->object = object;
    binding->func   = func;
    binding->domain = g_strdup (domain);
    binding->key    = g_strdup (key);

    /* FIXME: Create binding_object_destroy */ 
    closure = g_cclosure_new (G_CALLBACK(binding_object_signal_handler_invoked), binding, NULL);
    g_closure_set_marshal (closure, (GClosureMarshal)object_binding_marshal);
    g_signal_connect_closure ((gpointer)object, signal_spec, closure, TRUE);    
}

void
bmp_configuration_bind_toggle_button (BmpConfiguration	*self,
				      GtkToggleButton	*button,
				      const gchar	*domain,
				      const gchar	*key)
{
    g_return_if_fail (BMP_IS_CONFIGURATION(self));

    bmp_configuration_bind_object (self,
				   G_OBJECT(button),
				   toggle_button_binding,
				   "toggled",
				   domain, 
				   key); 
}

void
bmp_configuration_bind_toggle_action (BmpConfiguration	*self,
				      GtkToggleAction	*action,
				      const gchar	*domain,
				      const gchar	*key)
{
    g_return_if_fail (BMP_IS_CONFIGURATION(self));

    bmp_configuration_bind_object (self,
				   G_OBJECT(action),
				   toggle_action_binding,
				   "toggled",
				   domain, 
				   key); 
}
#endif

gboolean
bmp_configuration_save (BmpConfiguration * configuration)
{
    gchar *filename;
    FILE *file_xml;

    filename =
        g_build_filename (g_get_home_dir(), BMP_RCPATH, BMP_CONFIG_BASENAME, NULL);
    file_xml = g_fopen (filename, "w+");
    g_free (filename);

    xmlDocFormatDump (file_xml,
                      configuration->private->xmlconfig, 1);

    fclose (file_xml);

    return TRUE;
}

static GObject *bmp_configuration_constructor (GType type,
                                             guint n_construct_properties,
                                             GObjectConstructParam *
                                             construct_properties);

static void
bmp_configuration_init (BmpConfiguration * configuration)
{
    xmlNodePtr node;
    gchar *filename;

    configuration->private = g_new (BmpConfigurationPrivate, 1);

    configuration->private->dispose_has_run = FALSE;

    filename = g_build_filename (g_get_home_dir(), BMP_RCPATH, NULL);

    /* FIXME: As not only the configuration is stored there, this shouldn't be the
       function to create the BMP_RCPATH directory */

    if (!g_file_test (filename, G_FILE_TEST_EXISTS))
        mkdir (filename, S_IRWXU);

    g_free (filename);

    filename = g_build_filename (g_get_home_dir(), BMP_RCPATH, "config.xml", NULL);

    if (g_file_test (filename, G_FILE_TEST_EXISTS))
    {
         configuration->private->xmlconfig =
         xmlParseFile (filename);
    }
    else
    {
         configuration->private->xmlconfig = xmlNewDoc ((unsigned char*)"1.0");

         node = xmlNewNode (NULL, (unsigned char*)ROOT_NODE_NAME);
         xmlSetProp (node, (unsigned char*)"version", (unsigned char*)VERSION);

         xmlAddChild (node, xmlNewText ((unsigned char*)"\n"));

         xmlDocSetRootElement (configuration->private->
                               xmlconfig, node);

	 bmp_configuration_key_set_string (configuration, "bmp","font", BMP_DEFAULT_FONT);
	 bmp_configuration_key_set_string (configuration, "bmp","playlists-path", BMP_PATH_PLAYLISTS);
	 bmp_configuration_key_set_string (configuration, "bmp","titlestring-format","<artist> - <album> - <title>");
	 bmp_configuration_key_set_string (configuration, "bmp","icon-theme","red");
	 bmp_configuration_key_set_string (configuration, "bmp","file-chooser-path", g_get_home_dir());
         bmp_configuration_key_set_string (configuration, "bmp","skin", 
	  	DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN);

	 bmp_configuration_key_set_boolean (configuration, "bmp","use-custom-cursors", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","display-tooltips", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","display-tracklist-numbers", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","font-per-skin",FALSE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","use-custom-cursors", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","keep-above", FALSE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","display-tracklist-numbers", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","shuffle", FALSE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","repeat", FALSE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","file-chooser-close-on-open", TRUE);
	 bmp_configuration_key_set_boolean (configuration, "bmp","file-chooser-close-on-add", FALSE);

#ifdef HAVE_GUI
	 bmp_configuration_key_set_int (configuration, "playlist-window","vsep-width", VSEP_MIN_WIDTH);
#endif

    }

    g_free (filename);

    g_message("%s\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(configuration));
}

static xmlXPathObjectPtr
bmp_configuration_get_from_xpath (BmpConfiguration * configuration,
                                  const gchar * xpath_expr)
{
    xmlXPathObjectPtr xpathObj = NULL;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), NULL);

    xpathObj =
        xml_execute_xpath_expression ((xmlDocPtr)configuration->private->
                                         xmlconfig,
                                         (unsigned char*)xpath_expr, NULL);
    return xpathObj;
}

gboolean
bmp_configuration_domain_new (BmpConfiguration * configuration,
                            const gchar * domain_id)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);

    if (bmp_configuration_domain_exists (configuration, domain_id))
      {
        return TRUE;
      }

    xpath_expr = 
      g_strdup_printf ("/%s",ROOT_NODE_NAME);
    xpathObj = 
      bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
      xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    node = xmlNewNode (NULL, (unsigned char*)"domain");

    xmlSetProp (node, (unsigned char*)"id", (unsigned char*)domain_id);
    xmlAddChild (nv->nodeTab[0], node);
    xmlAddChild (nv->nodeTab[0], xmlNewText ((unsigned char*)"\n\n"));
    xmlAddChild (node, xmlNewText ((unsigned char*)"\n"));

    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);


    return TRUE;
}

gboolean
bmp_configuration_key_new (BmpConfiguration *configuration,
                           const gchar	    *domain_id,
                           const gchar	    *key_id, 
			   GValue	    *key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gchar *sigdetail;
    gchar *key_value_str = NULL;
    const gchar *type_id = NULL;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail (domain_id != NULL, FALSE);

    xpath_expr = 
      g_strdup_printf ("/%s/domain[@id = '%s']", ROOT_NODE_NAME, domain_id);
    xpathObj = 
      bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
      xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (G_VALUE_HOLDS_STRING (key_value))
    {
        key_value_str = g_value_dup_string (key_value);
	type_id = "string";
    }
    else if (G_VALUE_HOLDS_INT (key_value))
    {
        key_value_str = g_strdup_printf ("%d", g_value_get_int (key_value));
	type_id = "integer";
    }
    else if (G_VALUE_HOLDS_BOOLEAN (key_value))
    {
	key_value_str  = g_value_get_boolean (key_value) ? g_strdup("TRUE") : g_strdup("FALSE");
	type_id = "boolean";
    }

    node =
      xmlNewTextChild (nv->nodeTab[0], NULL, (unsigned char*)"key", (unsigned char*)key_value_str);
    g_free (key_value_str);

    xmlAddSibling (node, xmlNewText ((unsigned char*)"\n"));

    if (key_id)
     {
         xmlSetProp (node, (unsigned char*)"id", (unsigned char*)key_id);
         xmlSetProp (node, (unsigned char*)"type", (unsigned char*)type_id);
         sigdetail = g_strdup_printf ("%s-%s", domain_id, key_id);
     }
    else
     {
         sigdetail = g_strdup (domain_id);
     }

    g_signal_emit (G_OBJECT (configuration),
                   signals[BMP_CONFIGURATION_SIGNAL_ADDED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);

    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);
    return TRUE;
}

/* This fetches all keys of a given domain */
GHashTable *
bmp_configuration_domain_fetch (BmpConfiguration * configuration,
                                         const gchar * domain_id)
{
    GHashTable *ht;
    xmlNodeSetPtr nv;
    xmlNodePtr node;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *key_value, *key_id;
    gint n;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), NULL);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key", ROOT_NODE_NAME, domain_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    ht = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

    for (n = 0; n < nv->nodeNr; ++n)
     {
         node = nv->nodeTab[n];

         if (!node->children)
             continue;

         key_value = (char*)XML_GET_CONTENT (node->children);

         /* _replace() will overwrite keys with the same id; but since
          * it's an id, it should be duplicate anyhow. Still i think
          */

         if (!xmlGetProp (node, (unsigned char*)"id"))
          {
              key_id = g_strdup_printf ("%d", n);
          }
         else
          {
              key_id = g_strdup ((char*)xmlGetProp (node, (unsigned char*)"id"));
          }

         g_hash_table_replace (ht, key_id, g_strdup (key_value));
     }

    xmlXPathFreeObject (xpathObj);
    return ht;
}

void
bmp_configuration_ht_destroy (BmpConfiguration *configuration,
                              GHashTable *ht)
{
    g_return_if_fail (BMP_IS_CONFIGURATION (configuration));
    g_hash_table_destroy (ht);
}

gboolean
bmp_configuration_key_unset (BmpConfiguration * configuration,
                         const gchar * domain_id,
                         const gchar * key_id,
                         gboolean create)
{
    xmlNodeSetPtr nv;
    xmlNodePtr node_cur;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
    xpathObj =
        bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv =
	xpathObj->nodesetval;  

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }


    if (!nv->nodeNr)
    {
	xmlXPathFreeObject (xpathObj);
        return bmp_configuration_key_new (configuration,
                                          domain_id,
                                          key_id,
                                          NULL);
    }

    node_cur =
        nv->nodeTab[0];

    if (node_cur->children)
    {
         xmlUnlinkNode (node_cur->children);
         xmlFreeNode (node_cur->children);
    }


    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);
    return TRUE;
}

gboolean
bmp_configuration_key_set (BmpConfiguration *configuration,
                           const gchar	    *domain_id,
                           const gchar	    *key_id,
                           GValue	    *key_value, 
			   gboolean	     create)
{
    xmlNodeSetPtr nv;
    xmlNodePtr node_cur;
    xmlNodePtr node_new;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *key_value_str = NULL, *sigdetail;
    const gchar *type_id = NULL;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);

    if (!bmp_configuration_domain_exists (configuration, domain_id))
      {
  	 bmp_configuration_domain_new (configuration, domain_id);
      }

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);	
    if (!xpathObj) return FALSE;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;

    if (!nv)
     {
         if (create)
          {
	      xmlXPathFreeObject (xpathObj);
              return bmp_configuration_key_new (configuration,
                                                domain_id, key_id, key_value);
          }
         else
	  {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	  }
     }

    if (!nv->nodeNr)
     {
         if (create)
          {
	      xmlXPathFreeObject (xpathObj);
              return bmp_configuration_key_new (configuration,
                                                domain_id, key_id, key_value);
          }
         else
	  {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	  }
     }

    if (G_VALUE_HOLDS_STRING (key_value))
     {
        key_value_str =
             g_value_dup_string (key_value);
	type_id = "string";
     }
    else
    if (G_VALUE_HOLDS_INT (key_value))
     {
        key_value_str =
             g_strdup_printf ("%d", g_value_get_int (key_value));
	type_id = "integer";
     }
     else
     if (G_VALUE_HOLDS_BOOLEAN (key_value))
     {
        key_value_str =
             g_strdup_printf ("%s", g_value_get_boolean (key_value) ? "TRUE" : "FALSE");
	type_id = "boolean";
     }

    node_cur =
        nv->nodeTab[0];
    node_new =
        xmlNewText ((unsigned char*)key_value_str);

    if (node_cur->children)
         xmlReplaceNode (node_cur->children, node_new);
    else
         xmlAddChild (node_cur, node_new);

    xmlSetProp(node_cur,(unsigned char*)"type",(unsigned char*)type_id);

    sigdetail = g_strdup_printf ("%s-%s", domain_id,key_id);

    g_signal_emit (G_OBJECT (configuration),
                   signals[BMP_CONFIGURATION_SIGNAL_CHANGED],
                   g_quark_from_string (sigdetail), NULL);

    g_free (sigdetail);
    g_free (key_value_str);
    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);
    return TRUE;
}

gint
bmp_configuration_key_fetch_int (BmpConfiguration * configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *aux;
    gint val = -1;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), -1);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj) return -1;
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return -1;
      }

    if (nv->nodeTab[0]->children)
    {
       aux =
            (char*)XML_GET_CONTENT((xmlNodePtr) (nv->nodeTab[0])->children);
       val =
	    strtol(aux,NULL,10);
    }

    xmlXPathFreeObject (xpathObj);
    return val;
}

gboolean
bmp_configuration_key_fetch_boolean (BmpConfiguration * configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *aux;
    gboolean val = FALSE;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj) return FALSE;
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (nv->nodeTab[0]->children)
    {
	aux = g_strdup((char*)XML_GET_CONTENT((xmlNodePtr)(nv->nodeTab[0]->children)));
	g_strstrip(aux);
	if (g_ascii_strcasecmp(aux, "TRUE") == 0)
		val = TRUE;
	if (g_ascii_strcasecmp(aux, "1") == 0)
		val = TRUE;
	g_free (aux);
    }

    xmlXPathFreeObject (xpathObj);
    return val;
}

const gchar*
bmp_configuration_key_fetch (BmpConfiguration * configuration,
                                 const gchar * domain_id,
                                 const gchar * key_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), NULL);

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj) return NULL;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    if (nv->nodeTab[0]->children)
      {
	gchar *rv;
        rv =
            (char*)XML_GET_CONTENT
                      ((xmlNodePtr) (nv->nodeTab[0])->children);
	xmlXPathFreeObject (xpathObj);
	return rv;
      } 
    else
      {
	xmlXPathFreeObject (xpathObj);
        return NULL;
      }

    xmlXPathFreeObject (xpathObj);
    return NULL;
}

gboolean
bmp_configuration_domain_exists (BmpConfiguration * configuration,
                               const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;

    xpath_expr = 
	g_strdup_printf ("/%s/domain[@id = '%s']", ROOT_NODE_NAME, domain_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);

    if (!xpathObj)
        return FALSE;

    nv = 
	xpathObj->nodesetval;

    if (!nv)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    if (!nv->nodeNr)
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      }

    xmlXPathFreeObject (xpathObj);
    return TRUE;
}

gboolean
bmp_configuration_domain_has_keys (BmpConfiguration * configuration,
                                 const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gboolean key_found = FALSE;

    xpath_expr =
        g_strdup_printf ("/%s/domain[@id = '%s']/key", ROOT_NODE_NAME, domain_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    g_free (xpath_expr);
    if (!xpathObj)
      return FALSE;
    nv = 
	xpathObj->nodesetval;	

    if (!nv)
        return FALSE;

    if (!nv->nodeNr)
        key_found = FALSE;
    else
        key_found = TRUE;

    xmlXPathFreeObject (xpathObj);
    return key_found;
}

gboolean
bmp_configuration_key_find (BmpConfiguration * configuration,
                          const gchar * domain_id,
                          const gchar * key_id, const gchar * key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr;
    gchar *key_value_cmp = NULL;
    gint n;

    gboolean key_found = FALSE;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);

    if (key_id)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
         xpathObj = 
	      bmp_configuration_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
	    return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

	 key_found = !!nv;
    
	 xmlXPathFreeObject (xpathObj);
         return key_found;
     }
    else if (key_value)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key", ROOT_NODE_NAME, domain_id);
         xpathObj = 
	     bmp_configuration_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
	    return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         for (n = 0; n < nv->nodeNr; ++n)
          {
              node = nv->nodeTab[n];

	      if (!key_value_cmp) continue;

              key_value_cmp = (char*)XML_GET_CONTENT (node->children);

              if (strcmp (key_value_cmp, key_value) == 0)
               {
                   key_found = TRUE;
                   break;
               }
          }
	 xmlXPathFreeObject (xpathObj);
         return key_found;
     }

    g_assert_not_reached ();
}

gboolean
bmp_configuration_remove_domain (BmpConfiguration * configuration,
                               const gchar * domain_id)
{
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;
    gchar *xpath_expr, *sigdetail;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);

    xpath_expr = 
	g_strdup_printf ("/%s/domain[@id = '%s']", ROOT_NODE_NAME, domain_id);
    xpathObj = 
	bmp_configuration_get_from_xpath (configuration, xpath_expr);
    if (!xpathObj)
      return FALSE;
    g_free (xpath_expr);
    nv = 
	xpathObj->nodesetval;	

    if ((!nv) || (!nv->nodeNr))
      {
	xmlXPathFreeObject (xpathObj);
        return FALSE;
      } 

    xmlUnlinkNode (nv->nodeTab[0]);
    xmlFreeNode (nv->nodeTab[0]);

    sigdetail = g_strdup_printf ("%s", domain_id);
    g_signal_emit (G_OBJECT (configuration),
                   signals[BMP_CONFIGURATION_SIGNAL_REMOVED],
                   g_quark_from_string (sigdetail), NULL);
    g_free (sigdetail);

    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);
    return TRUE;
}

gboolean
bmp_configuration_remove_key (BmpConfiguration * configuration,
                            const gchar * domain_id,
                            const gchar * key_id, const gchar * key_value)
{
    xmlNodePtr node;
    xmlNodeSetPtr nv;
    xmlXPathObjectPtr xpathObj;

    gchar *xpath_expr, *key_value_cmp, *sigdetail;
    gint n;

    g_return_val_if_fail (BMP_IS_CONFIGURATION (configuration), FALSE);
    g_return_val_if_fail ((domain_id != NULL), FALSE);
    g_return_val_if_fail ((key_value != NULL), FALSE);

    if (key_id)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key[@id = '%s']", ROOT_NODE_NAME, domain_id, key_id);
         xpathObj = 
	     bmp_configuration_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
            return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         xmlUnlinkNode (nv->nodeTab[0]);
         xmlFreeNode (nv->nodeTab[0]);

         sigdetail = g_strdup_printf ("%s-%s", domain_id, key_id);
         g_signal_emit (G_OBJECT (configuration),
                        signals[BMP_CONFIGURATION_SIGNAL_REMOVED],
                        g_quark_from_string (sigdetail), NULL);
         g_free (sigdetail);
	 xmlXPathFreeObject (xpathObj);
     }
    else if (key_value)
     {
         xpath_expr =
             g_strdup_printf ("/%s/domain[@id = '%s']/key", ROOT_NODE_NAME, domain_id);
         xpathObj = 
	     bmp_configuration_get_from_xpath (configuration, xpath_expr);
         if (!xpathObj)
            return FALSE;
         g_free (xpath_expr);
         nv = 
	     xpathObj->nodesetval; 

         if ((!nv) || (!nv->nodeNr))
	   {
	     xmlXPathFreeObject (xpathObj);
             return FALSE;
	   }

         /*
          * Unlink node here
          */
         for (n = 0; n < nv->nodeNr; ++n)
          {
              node = nv->nodeTab[n];

              key_value_cmp = (char*)XML_GET_CONTENT (node->children);

	      if (!key_value_cmp) continue;

              if (strcmp (key_value_cmp, key_value) == 0)
               {
                   xmlUnlinkNode (node);
                   xmlFreeNode (node);
               }
          }
         sigdetail = g_strdup_printf ("%s__key", domain_id);
         g_signal_emit (G_OBJECT (configuration),
                        signals[BMP_CONFIGURATION_SIGNAL_REMOVED],
                        g_quark_from_string (sigdetail), NULL);
         g_free (sigdetail);
     }

    xmlXPathFreeObject (xpathObj);
    bmp_configuration_save (configuration);
    return TRUE;
}


static void
bmp_configuration_set_property (GObject * object,
                              guint property_id, const GValue * value,
                              GParamSpec * pspec)
{
    switch (property_id)
     {
     }
}

static void
bmp_configuration_get_property (GObject * object,
                              guint property_id, GValue * value,
                              GParamSpec * pspec)
{
    switch (property_id)
     {
     }
}

static void
bmp_configuration_dispose (GObject * obj)
{
    BmpConfiguration *configuration = (BmpConfiguration *) obj;

    g_return_if_fail (!configuration->private->dispose_has_run);

    configuration->private->dispose_has_run = TRUE;

    g_message("%s\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_configuration_finalize (GObject * obj)
{
    BmpConfiguration *configuration = (BmpConfiguration *) obj;

    g_free (configuration->private);
}


BmpConfiguration *
bmp_configuration_new (void)
{
    BmpConfiguration *configuration;

    configuration = 
      g_object_new (BMP_TYPE_CONFIGURATION, NULL);

    return configuration;
}

static GObject *
bmp_configuration_constructor (GType type,
                             guint n_construct_properties,
                             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        BmpConfigurationClass *klass;
        GObjectClass *parent_class;
        klass =
            BMP_CONFIGURATION_CLASS (g_type_class_peek
                                        (BMP_TYPE_CONFIGURATION));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }
    return obj;
}

static void
bmp_configuration_class_init (BmpConfigurationClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);



    gobject_class->set_property = bmp_configuration_set_property;

    gobject_class->get_property = bmp_configuration_get_property;

    gobject_class->dispose = bmp_configuration_dispose;

    gobject_class->finalize = bmp_configuration_finalize;

    gobject_class->constructor = bmp_configuration_constructor;

    signals[BMP_CONFIGURATION_SIGNAL_REMOVED] =
        g_signal_new ("removed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpConfigurationClass, removed),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);

    signals[BMP_CONFIGURATION_SIGNAL_CHANGED] =
        g_signal_new ("changed",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpConfigurationClass, changed),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);


    signals[BMP_CONFIGURATION_SIGNAL_ADDED] =
        g_signal_new ("added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpConfigurationClass, added),
                      NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
                      0);

}

void
bmp_configuration_key_set_boolean (BmpConfiguration * configuration,
                         const gchar * domain_id,
                         const gchar * key_id,
                         gboolean value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_BOOLEAN);
    g_value_set_boolean (&value, value_arg);

    bmp_configuration_key_set (configuration, domain_id, key_id, &value, TRUE);
}

void
bmp_configuration_key_set_int (BmpConfiguration * configuration,
                         const gchar * domain_id,
                         const gchar * key_id,
                         gint value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_INT);
    g_value_set_int (&value, value_arg);

    bmp_configuration_key_set (configuration, domain_id, key_id, &value, TRUE);
}

void
bmp_configuration_key_set_string (BmpConfiguration * configuration,
                                  const gchar * domain_id,
                                  const gchar * key_id,
                                  const gchar *value_arg)
{
    GValue value = {0,};

    g_value_init (&value, G_TYPE_STRING);
    g_value_set_string (&value, value_arg);

    bmp_configuration_key_set (configuration, domain_id, key_id, &value, TRUE);
}
