/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id: main.c 336 2005-10-08 13:49:42Z mderezynski $
 */

#include <config.h>
#include <revision.h>
#include <build.h>

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#else
#  include <locale.h>
#endif

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>
#include <stdlib.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "main.h"
#include "error.h"


#ifdef HAVE_GUI
#  include "ui.h"
#  include "sm.h"
#  include "fam_monitor.h"
#endif


#include "configuration.h"
#include "gc.h"
#include "logger.h"
#include "play.h"
#include "resource_manager.h"
#include "signals.h"

#include "bmpx/sanity.h"
#include "bmpx/plugin.h"
#include "bmpx/vfs.h"
#include "bmpx/metadata.h"
#include "bmpx/playlist.h"

#include "plugin_priv.h"


#ifdef HAVE_GST
#  include <gst/gst.h>
#endif

#ifdef HAVE_DBUS
#  include <dbus/dbus-gtype-specialized.h>
#endif

#include "g_object_get.h"
#include "bmpx/util.h"

gchar *bmp_paths[BMP_N_PATHS] = { NULL };

BmpResourceManager *bmp_rm = NULL;
BmpConfiguration   *bmp_configuration = NULL;
BmpPlaylist        *bmp_playlist = NULL;
BmpPlay		   *bmp_play = NULL;
BmpSystemControl   *bmp_system_control = NULL;
FamMonitor	   *bmp_fam_monitor = NULL;
#ifdef HAVE_GUI
BmpUI              *bmp_ui = NULL;
#endif
BmpMetadataCache   *bmp_metadata_cache = NULL;
NGC		   *ngc = NULL;

#ifndef HAVE_GUI
GMainLoop	   *mainloop;
#endif

#ifdef HAVE_SM
static gchar *session_id = NULL;
#endif

static gboolean show_version  = FALSE;
#ifdef HAVE_GUI
static gboolean run_headless  = FALSE;
#endif
static gboolean become_daemon = FALSE;
static gboolean log_stdout    = FALSE;

#ifdef HAVE_DBUS
static gboolean no_remote     = FALSE;
#endif

static GOptionEntry bmp_options[] =
  {
    {"version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
     N_("Display version"), NULL},
    {"no-log", 0, 0, G_OPTION_ARG_NONE, &log_stdout,
     N_("Disable logfile, log messages to shell"), NULL},
#ifdef HAVE_GUI
    {"no-ui", 0, 0, G_OPTION_ARG_NONE, &run_headless,
     N_("Disable UI"), NULL},
    {"daemon", 0, 0, G_OPTION_ARG_NONE, &become_daemon,
     N_("Run as a daemon (implies --no-ui)"), NULL},
#else
    {"daemon", 0, 0, G_OPTION_ARG_NONE, &become_daemon,
     N_("Run as a daemon"), NULL},
#endif

#ifdef HAVE_DBUS
    {"no-remote", 0, 0, G_OPTION_ARG_NONE, &no_remote,
     N_("Disable DBUS remote interface"), NULL},
#endif
#ifdef HAVE_SM
    {"sm-client-id", 0, 0, G_OPTION_ARG_STRING, &session_id,
     N_("Specify session management ID"), N_("ID")},
#endif
    { NULL }
  };

static GList *
get_feature_list (void)
{
  GList *list = NULL;

#ifdef HAVE_SM
  list = g_list_append (list, N_("feature|X11 Session Management"));
#endif
#ifdef HAVE_DBUS
  list = g_list_append (list, N_("feature|D-BUS Support"));
#endif
#ifdef HAVE_XINE
  list = g_list_append (list, N_("feature|Audio backend: XINE"));
#endif
#ifdef HAVE_GST
  list = g_list_append (list, N_("feature|Audio backend: GStreamer"));
#endif
#ifdef HAVE_GUI
  list = g_list_append (list, N_("feature|Core GUI (Winamp 2.x skinning engine)"));
#endif

  return list;
}

static void
print_version (void)
{
  GList *features, *iter;

  g_print ("%s %s, Build: %s-R%s [%s] (%s)\n\n", PACKAGE, VERSION, RV_LAST_CHANGED_DATE, RV_REVISION, BUILD_DATE, BUILD_BUILDUSER);

  features = get_feature_list ();

  g_print (ngettext ("Feature built: \n", "Features built: \n",
                     g_list_length (features)));

  for (iter = features; iter; iter = g_list_next (iter))
    g_print ("* %s\n", Q_(iter->data));

  g_list_free (features);
}

static gboolean
make_user_dir (const gchar *name)
{
  const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
  return g_mkdir (name, mode755);
}

static void
bmp_make_user_dirs (void)
{
  make_user_dir (BMP_PATH_USER_DIR);
  make_user_dir (BMP_PATH_USER_SKIN_DIR);
  make_user_dir (BMP_PATH_SKIN_THUMB_DIR);
  make_user_dir (BMP_PATH_PLAYLISTS);
}

static void
bmp_free_paths (void)
{
  int i;

  for (i = 0; i < BMP_N_PATHS; i++)
    {
      g_free (bmp_paths[i]);
      bmp_paths[i] = NULL;
    }
}

#define USER_PATH(path) \
    g_build_filename (BMP_PATH_USER_DIR, path, NULL);

static void
bmp_init_paths (void)
{
  BMP_PATH_USER_DIR = g_build_filename(g_get_home_dir(), BMP_RCPATH, NULL);
  BMP_PATH_USER_SKIN_DIR = USER_PATH(BMP_SKIN_DIR_BASENAME);
  BMP_PATH_SKIN_THUMB_DIR = USER_PATH(BMP_SKIN_THUMB_DIR_BASENAME);
  BMP_PATH_PLAYLISTS = USER_PATH(BMP_PLAYLISTS_DIR_BASENAME);
  BMP_PATH_LOG_FILE = USER_PATH(BMP_LOG_BASENAME);
  BMP_PATH_METADATA_CACHE = USER_PATH(BMP_METADATA_CACHE_BASENAME);
#if 0
  BMP_PATH_CONFIG_FILE = USER_PATH(BMP_CONFIG_BASENAME);
  BMP_PATH_ACCEL_FILE = USER_PATH(BMP_ACCEL_BASENAME);
#endif

  bmp_make_user_dirs ();
}

static void
setup_i18n (void)
{
#ifdef HAVE_GUI
  gtk_set_locale ();
#else
  setlocale (LC_ALL, "");
#endif

  bindtextdomain (PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (PACKAGE, "UTF-8");
  textdomain (PACKAGE);
}

gint
main (gint    argc,
      gchar** argv)
{
  GError *error = NULL;
  gchar  *sanity_error = NULL;
  GOptionContext *oc_bmp;

  setup_i18n ();

  sanity_error = sanity_check_libs ();
  if (sanity_error)
    {
      g_printerr ("%s\n", sanity_error);
      g_free (sanity_error);

      exit (EXIT_FAILURE);
    }

  /* Attempt to run X11-less if we are --daemon or --no-ui */
  oc_bmp = g_option_context_new(" - Run BMPx");
  g_option_context_add_main_entries (oc_bmp, bmp_options, PACKAGE);

#ifdef HAVE_GUI
  g_option_context_add_group (oc_bmp, gtk_get_option_group (FALSE));
#endif
  if (!g_option_context_parse (oc_bmp, &argc, &argv, &error))
    {
      g_printerr (Q_("Invalid argument: %s\n"), error->message);
      g_error_free (error);
      exit (EXIT_FAILURE);
    }

  g_type_init();

#ifdef HAVE_GUI
  if (become_daemon)
   {
     run_headless = TRUE;
   }
  else if (!run_headless)
   {
     gtk_init (&argc, &argv);
   }
  else
   {
   }
#endif

  if (show_version)
    {
      print_version ();
      return EXIT_SUCCESS;
    }

  ngc = gc_new ();
  vfs_init ();

#ifdef HAVE_GST
  gst_init (&argc, &argv);
#else
  g_thread_init (NULL);
#endif

  bmp_init_paths ();

  if (!log_stdout)
    {
      bmp_logger_start (BMP_PATH_LOG_FILE);
    }
  else
    {
      g_message ("BMPX STARTED [%s] PID [%d] ", get_timestamp_str (), getpid ());
    }

  /* Initialize Plugins */
  plugins_initialize();

  /* FIXME: gtk_init_with_args() (above) will remove all recognized
     arguments, so the session manager will run BMPx with incomplete
     options when it restarts. */
#ifdef HAVE_GUI
  if (!run_headless)
    session_start (argc, argv, session_id);
#endif

  bmp_rm = bmp_resource_manager_new ();
  bmp_configuration = bmp_configuration_new ();

  if (become_daemon)
    {
      if (!bmp_become_daemon ())
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unable to daemonize, proceeding with normal startup.");
#ifdef HAVE_GUI
      else
        run_headless = TRUE;
#endif
    }

#ifdef HAVE_GUI
  bmp_fam_monitor = fam_monitor_new ();
#endif

#ifdef HAVE_DBUS
  if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "no-remote"))
    no_remote = TRUE;
#endif

  bmp_playlist = bmp_playlist_new ();
  bmp_play = bmp_play_new ();

#ifdef HAVE_DBUS
  if (!no_remote)
    dbus_g_type_specialized_init ();

  bmp_system_control = bmp_system_control_new (no_remote);
#else
  bmp_system_control = bmp_system_control_new (FALSE);
#endif

  if (bmp_system_control)
    {
      signal_handlers_init ();

      bmp_metadata_cache = bmp_metadata_cache_new ();

      bmp_playlist_load_playlists (bmp_playlist);

#ifdef HAVE_GUI
      if (bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "no-ui"))
        run_headless = TRUE;

      if (!run_headless)
        bmp_system_control_ui_start (bmp_system_control);
      else
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Not starting UI on request.");
#endif

      bmp_system_control_send_startup_complete (bmp_system_control);
  
#ifdef HAVE_GUI
      gtk_main ();
#else
      mainloop = g_main_loop_new (NULL, FALSE);
      g_main_loop_run (mainloop);
#endif
    }
  else
    {
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "BmpSystemControl not present, cannot proceed with startup sequence.");
    }

  vfs_stop();

  if ( bmp_play && bmp_system_control)
    bmp_play_stop_stream (bmp_play);

#ifdef HAVE_GUI
  if (bmp_fam_monitor)
    {
      g_object_unref (bmp_fam_monitor);
      bmp_fam_monitor = NULL;
    }

  if (bmp_ui)
    {
      g_object_unref (bmp_ui);
      bmp_ui = NULL;
    }
#endif

  if (bmp_metadata_cache)
    {
      g_object_unref (bmp_metadata_cache);
      bmp_metadata_cache = NULL;
    }

  if (bmp_play)
    {
      g_object_unref (bmp_play);
      bmp_play = NULL;
    }

  if (bmp_playlist)
    {
      g_object_unref (bmp_playlist);
      bmp_playlist = NULL;
    }

  if (bmp_configuration)
    {
      g_object_unref (bmp_configuration);
      bmp_configuration = NULL;
    }


  if (bmp_system_control)
    {
      bmp_system_control_send_shutdown_complete (bmp_system_control);
      g_object_unref (bmp_system_control);
      bmp_system_control = NULL;
    }

#ifdef HAVE_GUI
  if (!run_headless)
    session_end ();
#endif

  if (!log_stdout)
    {
      bmp_logger_stop ();
    }
  else
    {
      g_message (" ");
      g_message ("BMPX STOPPED [%s]", get_timestamp_str());
    }

  bmp_free_paths ();

  /* Shutdown Plugins */
  plugins_shutdown();

  gc_destroy (ngc);

  exit (EXIT_SUCCESS);
}
