#include <config.h>

#ifdef HAVE_GUI
#include <gtk/gtk.h>
#else
#include <glib.h>
#endif

#include <glib/gstdio.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "main.h"

#include <taglib/tag_c.h>

#include <libhrel/types.h>
#include <libhrel/relation.h>

#include "bmpx/util.h"
#include "bmpx/file_utils.h"
#include "bmpx/metadata.h"
#include "bmpx/uri.h"


G_DEFINE_TYPE (BmpMetadataCache, bmp_metadata_cache, G_TYPE_OBJECT)

enum {
    BMP_METADATA_CACHE_SIGNAL_FILE_ADDED,

    BMP_METADATA_N_SIGNALS
};
static guint signals[BMP_METADATA_N_SIGNALS] = {0};

enum {
    LIST_ARTIST,
    LIST_ALBUM
};

struct _BmpMetadataCachePrivate {
        gboolean dispose_has_run;

	FILE *fh;

	HRelation   *relation;
};

struct _BmpTagDefine {
    gchar *title;
    gchar *id;
    GType  type;

    gint   default_int;
    gchar* default_char;
};
typedef struct _BmpTagDefine BmpTagDefine;

/* We're using GST-style tag identifiers here */
static BmpTagDefine bmp_tag_defines[] = {
    {"Location",          "location",       G_TYPE_STRING, 0, "(Unknown Location)"},
    {"Title",             "title",          G_TYPE_STRING, 0, "(Unknown Title)"},
    {"Artist",            "artist",         G_TYPE_STRING, 0, "(Unknown Artist)"},
    {"Album",             "album",          G_TYPE_STRING, 0, "(Unknown Album)"},
    {"Genre",             "genre",          G_TYPE_STRING, 0, "(Unknown Genre)"},
    {"Comment",           "comment",        G_TYPE_STRING, 0, "(Unknown Comment)"},

    {"Time",              "time",           G_TYPE_INT,	   0, ""},
    {"Track",             "track-number",   G_TYPE_INT,	   0, ""},
    {"Rating",            "rating",         G_TYPE_INT,	   0, ""},
    {"Date",              "date",           G_TYPE_INT,	   0, ""},
    {"MTIME",		  "mtime",	    G_TYPE_INT,	   0, ""},
    {"Bitrate",           "bitrate",        G_TYPE_INT,	   0, ""},
    {"Samplerate",        "samplerate",     G_TYPE_INT,	   0, ""},
    {"Play count",        "count",	    G_TYPE_INT,	   0, ""},
};

G_LOCK_DEFINE(METADATA);

static void
tracks_add_track (HRelationBase *relation, Track *track)
{
  h_relation_base_insert_tuple (relation,
                              bmp_metadata_get_id_static (BMP_DATUM_LOCATION),	  track->location,
                              bmp_metadata_get_id_static (BMP_DATUM_TITLE),	  track->title,
                              bmp_metadata_get_id_static (BMP_DATUM_ARTIST),	  track->artist,
                              bmp_metadata_get_id_static (BMP_DATUM_ALBUM),	  track->album,
                              bmp_metadata_get_id_static (BMP_DATUM_GENRE),	  track->genre,
                              bmp_metadata_get_id_static (BMP_DATUM_COMMENT),	  track->comment,
			      bmp_metadata_get_id_static (BMP_DATUM_TIME),	  track->time,
                              bmp_metadata_get_id_static (BMP_DATUM_TRACK),	  track->track,
                              bmp_metadata_get_id_static (BMP_DATUM_RATING),	  track->rating,
                              bmp_metadata_get_id_static (BMP_DATUM_DATE),	  track->date,
                              bmp_metadata_get_id_static (BMP_DATUM_MTIME),	  track->mtime,
                              bmp_metadata_get_id_static (BMP_DATUM_BITRATE),	  track->bitrate,
                              bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE),  track->samplerate,
                              bmp_metadata_get_id_static (BMP_DATUM_COUNT),	  track->count);
}

static void
tracks_delete_track_with_uri (HRelationBase *relation, const gchar *uri)
{
  h_relation_base_delete_tuple (relation, uri);
}

static GHashTable*
tuple_to_hash_table (HTuple *tuple)
{
  GHashTable *array;
  GValue *value;
  HTupleElem *elmt;
  const gchar *key_name;
  gint n;

  array = g_hash_table_new (g_str_hash, g_str_equal); 

  for (n = 0; n < G_N_ELEMENTS(bmp_tag_defines); n++)
    {
      key_name = 
	  bmp_metadata_get_id_static (n); 
      elmt = 
	  h_tuple_get (tuple, key_name); 

      value = g_new0 (GValue,1);
      g_value_init (value, elmt->type);

      if (elmt->type == G_TYPE_INT)
  	g_value_set_int (value, GPOINTER_TO_INT(elmt->value));
      else if (elmt->type == G_TYPE_STRING)
	g_value_set_string (value, (gchar*)(elmt->value));

      g_hash_table_insert (array, g_strdup(key_name), value); 
    }

  return array;
}

GHashTable*
bmp_metadata_cache_get_hash_value_table (BmpMetadataCache *self, const gchar *uri)
{
  HTuple *tuple;

  tuple = bmp_metadata_cache_get_metadata (self, uri);

  if (!tuple)
    return NULL;

  return tuple_to_hash_table (tuple);
}

static Track*
tracks_get_track_from_uri (HRelationBase *relation, const gchar *uri)
{
  HTuple *tuple;

  tuple = h_relation_base_get_tuple (relation, uri);

  if (tuple)
    {
      /* marshal tuple into a Track */

      Track *track = g_new (Track, 1);

      track->location	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_LOCATION));   
      track->title	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_TITLE));   
      track->artist	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_ARTIST));   
      track->album	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_ALBUM));   
      track->genre	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_GENRE));   
      track->comment	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_COMMENT));   
      track->artist	= (gchar *) h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_ARTIST));   
      track->track	= (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_TRACK));   
      track->rating	= (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_RATING));   
      track->date	= (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_DATE));   
      track->mtime	= (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_MTIME));   
      track->bitrate    = (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_BITRATE));   
      track->samplerate = (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE));   
      track->count      = (gint)    h_tuple_get_value (tuple,  bmp_metadata_get_id_static (BMP_DATUM_COUNT));   

      h_tuple_free (tuple);

      return track;
    }

  return NULL;
}

gchar*
bmp_metadata_get_id (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].id);
}

GType
bmp_metadata_get_type (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].type;
}


const gchar*
bmp_metadata_get_id_static (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].id;
}


gchar*
bmp_metadata_get_title (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].title);
}

gchar*
bmp_metadata_get_default_char (BmpMetadatumId tag_id)
{
    return g_strdup(bmp_tag_defines[tag_id].default_char);
}

gint
bmp_metadata_get_default_int (BmpMetadatumId tag_id)
{
    return bmp_tag_defines[tag_id].default_int;
}




static Track*
metadata_get (const gchar *uri)
{
    TagLib_File *file = NULL;
    TagLib_Tag  *tag  = NULL;
    TagLib_AudioProperties const *audio = NULL;
    Track *track;
    GURI *g_uri;
    struct stat file_stat;
    gchar *filename;

    g_uri = bmp_uri_new (uri);
    bmp_uri_unescape (g_uri);
    filename = g_uri->path;

    if (!filename)
      {
	bmp_uri_delete (g_uri);
	return NULL;
      }

    track = g_new0 (Track,1);

    stat (filename, &file_stat);
    track->mtime    = file_stat.st_mtime;
    track->location = g_strdup (uri);

    file = taglib_file_new (filename);

    if (file)
      {
         tag = taglib_file_tag (file);
         if (!tag)
           {
              g_message ("%s: no metadata", filename);
	      taglib_file_free (file);
              return NULL;
           }

         audio = taglib_file_audioproperties (file);
	 track->track      = taglib_tag_track (tag);
	 track->date       = taglib_tag_year  (tag); 
	 track->artist     = g_strdup (taglib_tag_artist  (tag) ? taglib_tag_artist  (tag) : "(Unknown Artist)");
	 track->album      = g_strdup (taglib_tag_album   (tag) ? taglib_tag_album   (tag) : "(Unknown Album)");
	 track->title      = g_strdup (taglib_tag_title   (tag) ? taglib_tag_title   (tag) : "(Unknown Title)");
	 track->genre      = g_strdup (taglib_tag_genre   (tag) ? taglib_tag_genre   (tag) : "(Unknown Genre)");
	 track->comment    = g_strdup (taglib_tag_comment (tag) ? taglib_tag_comment (tag) : "(No Comment)");
	 track->time       = taglib_audioproperties_length (audio);
	 track->rating	   = 0;
	 track->count	   = 0;
         track->bitrate    = taglib_audioproperties_bitrate (audio);
         track->samplerate = taglib_audioproperties_samplerate (audio);
             
         taglib_tag_free_strings ();
         taglib_file_free (file);
     }
   else
     { 
	g_free (track->location);
	g_free (track);
	bmp_uri_delete (g_uri);
	return NULL;
     }

    bmp_uri_delete (g_uri);
    return track;
}

/* Media Cache functions */
static void
bmp_metadata_cache_init (BmpMetadataCache *self)
{
        self->private = g_new (BmpMetadataCachePrivate, 1);

        self->private->dispose_has_run = FALSE; 

	h_type_init ();

	self->private->relation = 
	    h_relation_base_new ("track",
                               "track has location, title, artist, album, genre, comment, track, rating, date, mtime, bitrate, samplerate, count",
                               bmp_metadata_get_id_static(BMP_DATUM_LOCATION),	  bmp_metadata_get_type (BMP_DATUM_LOCATION),
                               bmp_metadata_get_id_static(BMP_DATUM_TITLE),	  bmp_metadata_get_type (BMP_DATUM_TITLE),
                               bmp_metadata_get_id_static(BMP_DATUM_ARTIST),	  bmp_metadata_get_type (BMP_DATUM_ARTIST),
                               bmp_metadata_get_id_static(BMP_DATUM_ALBUM),	  bmp_metadata_get_type (BMP_DATUM_ALBUM),
                               bmp_metadata_get_id_static(BMP_DATUM_GENRE),	  bmp_metadata_get_type (BMP_DATUM_GENRE),
                               bmp_metadata_get_id_static(BMP_DATUM_COMMENT),	  bmp_metadata_get_type (BMP_DATUM_COMMENT),
                               bmp_metadata_get_id_static(BMP_DATUM_TIME),	  bmp_metadata_get_type (BMP_DATUM_TIME),
                               bmp_metadata_get_id_static(BMP_DATUM_TRACK),	  bmp_metadata_get_type (BMP_DATUM_TRACK),
                               bmp_metadata_get_id_static(BMP_DATUM_RATING),	  bmp_metadata_get_type (BMP_DATUM_RATING),
                               bmp_metadata_get_id_static(BMP_DATUM_DATE),	  bmp_metadata_get_type (BMP_DATUM_DATE),
                               bmp_metadata_get_id_static(BMP_DATUM_MTIME),	  bmp_metadata_get_type (BMP_DATUM_MTIME),
                               bmp_metadata_get_id_static(BMP_DATUM_BITRATE),     bmp_metadata_get_type (BMP_DATUM_BITRATE),
                               bmp_metadata_get_id_static(BMP_DATUM_SAMPLERATE),  bmp_metadata_get_type (BMP_DATUM_SAMPLERATE),
                               bmp_metadata_get_id_static(BMP_DATUM_COUNT),	  bmp_metadata_get_type (BMP_DATUM_COUNT),
                               NULL,
                               bmp_metadata_get_id_static(BMP_DATUM_LOCATION), 
                               NULL);

	g_message("%s\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static void
bmp_metadata_cache_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        switch (property_id) {
        default:
                g_assert (FALSE);
                break;
        }
}

static GObject *
bmp_metadata_cache_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                BmpMetadataCacheClass *klass;
                GObjectClass *parent_class;
                klass = BMP_METADATA_CACHE_CLASS (g_type_class_peek (BMP_TYPE_METADATA_CACHE));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
bmp_metadata_cache_dispose (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        if (self->private->dispose_has_run)
	  return;

        self->private->dispose_has_run = TRUE;

	bmp_metadata_cache_write (self);
	g_message("%s\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_metadata_cache_finalize (GObject *obj)
{
        BmpMetadataCache *self = (BmpMetadataCache *)obj;

        g_free (self->private);
}

BmpMetadataCache*
bmp_metadata_cache_new (void)
{
    BmpMetadataCache *self;

    self = g_object_new (bmp_metadata_cache_get_type (), NULL);
    bmp_metadata_cache_read (self);

    return self;
}

static void
bmp_metadata_cache_class_init (BmpMetadataCacheClass *g_class)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

        gobject_class->set_property = bmp_metadata_cache_set_property;
        gobject_class->get_property = bmp_metadata_cache_get_property;
        gobject_class->dispose = bmp_metadata_cache_dispose;
        gobject_class->finalize = bmp_metadata_cache_finalize;
        gobject_class->constructor = bmp_metadata_cache_constructor;

        signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED] =
	    g_signal_new ("file_added",
                      G_OBJECT_CLASS_TYPE (gobject_class),
                      G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
                      G_STRUCT_OFFSET (BmpMetadataCacheClass, file_added),
                      NULL, NULL, g_cclosure_marshal_VOID__INT,
                      G_TYPE_NONE, 1, G_TYPE_INT);

}

/* Public API */
HTuple*
bmp_metadata_cache_get_metadata (BmpMetadataCache *self, const gchar *uri)
{
    HTuple *tuple = NULL;
    Track *track = NULL;    

    G_LOCK(METADATA);

    tuple = h_relation_base_get_tuple (H_RELATION_BASE(self->private->relation), uri);
 
    if (!tuple)
    {
	if (!g_ascii_strncasecmp("file",uri,4))
	  {
		track = metadata_get (uri);
#ifdef HAVE_XINE
		if (!track)
		    track = bmp_play_get_metadata_xine (bmp_play, uri);
#endif
	  }
#ifdef HAVE_XINE
	else
		track = bmp_play_get_metadata_xine (bmp_play, uri);
#endif

	if (track)
	  {
	    tracks_add_track (H_RELATION_BASE(self->private->relation), track);
	    tuple = h_relation_base_get_tuple (H_RELATION_BASE(self->private->relation), uri);
	  }

	g_signal_emit (G_OBJECT(self), signals[BMP_METADATA_CACHE_SIGNAL_FILE_ADDED], 0, 0); 
     }


    G_UNLOCK(METADATA);

    return tuple; 
}

#define PACKET_LIBNAME   (uint8_t)0x01
#define PACKET_MEDIAFILE (uint8_t)0x02
#define PACKET_EMC       (uint8_t)0x03

#define PACKET_MF_MVALUE    (uint8_t)0x01
#define PACKET_MF_EMF       (uint8_t)0x02

static void
metadata_tuple_foreach (gpointer key, gpointer data, BmpMetadataCache *self)
{
    HTupleElem *elem;
    FILE *fh;
    uint8_t value;
    uint16_t value16;
    uint8_t packet_id;
    BmpMetadatumId tag_id = -1;
    gint n;
    gchar *aux;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    if (!data) return;

    for (n = 0; n < BMP_DATUM_NONE; n++)
      {
	if (!strcmp(key, bmp_metadata_get_id_static(n)))
	  {
	    tag_id = n;
	    break;
	  }
      }

    if (tag_id == -1) return;

    fh = self->private->fh;

    elem = (HTupleElem*)data;

    packet_id = PACKET_MF_MVALUE;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    value = tag_id;
    fwrite (&value, sizeof(uint8_t), 1, fh);
  
    if (elem->type == G_TYPE_INT)
      {
	aux = g_strdup_printf("%d", GPOINTER_TO_INT(elem->value));
      }
    else
      {
	aux = g_strdup (elem->value);
      }

    value16 = strlen(aux);
    fwrite (&value16, sizeof(uint16_t), 1, fh);
    fwrite (aux, sizeof(char), strlen(aux), fh);

    g_free (aux);  
}

static void
for_each_metadata_tuple (HTuple *tuple, BmpMetadataCache *self)
{
    FILE *fh;
    uint8_t  packet_id;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    fh = self->private->fh;

    /* Write MF identification */
    packet_id = PACKET_MEDIAFILE;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);

    /* Now write all metadata */
    g_hash_table_foreach (tuple, (GHFunc) metadata_tuple_foreach, self);

    /* Write EMF */
    packet_id = PACKET_MF_EMF;
    fwrite (&packet_id, sizeof(uint8_t), 1, fh);
}

void
bmp_metadata_cache_write (BmpMetadataCache *self)
{
    FILE *fh;
    static const gchar *mcid = "MC";
    static const uint16_t mcidn = 666;
    uint8_t value;
    gchar *path;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    /* We make the operation atomic by writing to a temp file first and only if we
       fully succeed we move the temp to the actual file */

    path = g_strconcat (BMP_PATH_METADATA_CACHE, ".tmp", NULL);

    self->private->fh = g_fopen(path, "w");
    fh = self->private->fh;

    /* Write file magic */
    fwrite (mcid, sizeof(char), strlen(mcid), fh);
    fwrite (&mcidn, sizeof(uint16_t), 1, fh);

    /* Process cache */
    h_relation_for_each_tuple (self->private->relation, (HTupleFunc) for_each_metadata_tuple, self);

    /* Write EMC */
    value = PACKET_EMC;
    fwrite (&value, sizeof(uint8_t), 1, fh);
    fclose (self->private->fh);

    g_remove (BMP_PATH_METADATA_CACHE);
    g_rename (path, BMP_PATH_METADATA_CACHE);
}

void
bmp_metadata_cache_read (BmpMetadataCache *self)
{
    FILE *fh;
    static gchar *mcid;
    static uint16_t mcidn;
    uint8_t packet_id;
    uint8_t value;
    uint16_t value16;
    char *valuestr;
    gboolean run = TRUE;

    g_return_if_fail (BMP_IS_METADATA_CACHE (self));

    if (!g_file_test(BMP_PATH_METADATA_CACHE, G_FILE_TEST_EXISTS)) return;

    self->private->fh = g_fopen(BMP_PATH_METADATA_CACHE, "r");
    fh = self->private->fh;

    /* Write file magic */
    mcid = g_malloc0(3);
    fread (mcid, sizeof(char), 2, fh);
    fread (&mcidn, sizeof(uint16_t), 1, fh);

    if (strcmp(mcid,"MC")) {
	g_message("%s: Wrong file magic of metadata cache file %s",
		  G_STRFUNC, BMP_PATH_METADATA_CACHE);
	fclose(self->private->fh);
	return;
    }

    if (mcidn != 666) {
	g_message("%s: Wrong file magic for metadata cache file %s",
		  G_STRFUNC, BMP_PATH_METADATA_CACHE);
	fclose(self->private->fh);
	return;
    }

    /* Read the metadata cache */
    while (run)
    {
	    fread (&packet_id, sizeof(uint8_t), 1, fh);

	    switch (packet_id)
	    {
		/* Currently unsupported */
		case PACKET_LIBNAME:
		    break;

		case PACKET_EMC:
		    {
			run = FALSE;
			break;
		    }

		case PACKET_MEDIAFILE:
		    {
			HTuple *tuple;
			gboolean run_mf = TRUE;
			gint n;

			/* Fill the tuple with default values */
			tuple = h_tuple_new (); 
			for (n = 0; n < G_N_ELEMENTS(bmp_tag_defines); n++)
			  {
			      GType type = bmp_metadata_get_type (n);

			      h_tuple_insert (tuple,
					      bmp_metadata_get_id_static (n), 
					      type, 
					      (type == G_TYPE_INT) ? GINT_TO_POINTER (bmp_metadata_get_default_int (n)) 
								   : bmp_metadata_get_default_char (n)); 
			  }

			while (run_mf)
			{
			    fread (&packet_id, sizeof(uint8_t), 1, fh);

			    switch (packet_id)
			    {
				case PACKET_MF_MVALUE:
				    {
					BmpMetadatumId tag_id;
					gint aux_int;

					fread (&value, sizeof(uint8_t), 1, fh);
					tag_id = value;

					fread (&value16, sizeof(uint16_t), 1, fh);
					valuestr = g_malloc0(value16+1);
					fread (valuestr, sizeof(char), value16, fh);

					if (bmp_metadata_get_type (tag_id) == G_TYPE_INT)
					  {
					      aux_int = strtol (valuestr, NULL, 10);
					      h_tuple_insert (tuple,
							      bmp_metadata_get_id_static (tag_id),
							      G_TYPE_INT,
							      GINT_TO_POINTER(aux_int));
					  }
					else
					  {
					      h_tuple_insert (tuple,
							      bmp_metadata_get_id_static (tag_id),
							      G_TYPE_STRING,
							      valuestr); 

					  }

					g_free (valuestr);

					break;
				    }

				case PACKET_MF_EMF:
				    {
					run_mf = FALSE;
					break;
				    }

				default:
				    {
					g_message("%s: Unknown packet type %d while reading media file record", G_STRFUNC, packet_id);
				    }
			    }
			}

#if 0
			if (!uri)
			  {
			    g_message("%s: packet has no URI, discarding", G_STRFUNC);
			    break;
			  }

			if (!g_ascii_strncasecmp("file",uri,4))
			  {
			    /* FIXME: Use GURI here */
			    filename =
				g_filename_from_uri (uri, NULL, NULL);

			    /* File is not present anymore, discard */
			    if (! g_file_test (filename, G_FILE_TEST_EXISTS)) {
				g_free (uri);
				g_free (filename);
				g_hash_table_destroy (metadata);
				break;
			    }
			    g_free (filename);
		          }

			  if (G_URI_IS_LOCAL(uri))
			    {
			    	struct stat file_stat;
				long int st_mtime_cmp;
				GURI *g_uri;

				st_mtime_cmp =
				    strtol (g_hash_table_lookup (metadata, bmp_metadata_get_id_static (BMP_DATUM_MTIME)), NULL, 10);
				g_uri =
				    bmp_uri_new (uri);

				stat (g_uri->path, &file_stat);

				if (st_mtime_cmp != file_stat.st_mtime)
			          {
				    g_hash_table_destroy (metadata);
				    g_free (uri);
				    bmp_uri_delete (g_uri);
				    break;
				  }

			        bmp_uri_delete (g_uri);
			    }
#endif
			h_relation_base_insert_tuple (H_RELATION_BASE(self->private->relation), 
                               bmp_metadata_get_id_static(BMP_DATUM_LOCATION),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_LOCATION))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_TITLE),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_TITLE))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_ARTIST),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_ARTIST))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_ALBUM),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_ALBUM))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_GENRE),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_GENRE))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_COMMENT),	  h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_COMMENT))->value,
                               bmp_metadata_get_id_static(BMP_DATUM_TIME),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_TIME))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_TRACK),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_TRACK))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_RATING),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_RATING))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_DATE),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_DATE))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_MTIME),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_MTIME))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_BITRATE),     GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_BITRATE))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_SAMPLERATE),  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_SAMPLERATE))->value),
                               bmp_metadata_get_id_static(BMP_DATUM_COUNT),	  GPOINTER_TO_INT(h_tuple_get(tuple, bmp_metadata_get_id_static (BMP_DATUM_COUNT))->value));

			h_tuple_free (tuple);
		      }
		    break;

		    default:
		      {
			g_message("%s: Unknown packet type '%d' while reading metadata cache", G_STRFUNC, packet_id);
			break;
		      }
	    }
    }

    fclose (self->private->fh);
}

void
bmp_metadata_cache_validate (BmpMetadataCache *self)
{
    g_return_if_fail (BMP_IS_METADATA_CACHE (self));
}

gchar*
bmp_metadata_cache_get_titlestring (BmpMetadataCache *self, const gchar *uri)
{
    GString *string_out = NULL;
    GString *string_field = NULL;
    HTuple *tuple;
    gchar *field;
    const gchar *format;
    gint n;
    gboolean hit = FALSE;

    g_return_val_if_fail (uri != NULL, NULL);
    
    tuple = bmp_metadata_cache_get_metadata (self, uri);

    if (!tuple)
      return g_strdup("");

    format = bmp_configuration_key_fetch (bmp_configuration, "bmp", "titlestring-format");
    string_out = g_string_new ("");

    while (*format) {

	/* Begin sequence */
	if (*format == '<') {

	    string_field =
		g_string_new ("");
	    format++;

	    while (*format && (*format != '>') && (*format != ' ')) {
		g_string_append_c (string_field, *format);
		format++;
	    }

	    field =
		g_string_free (string_field, FALSE);

	    if (!*format)
	      {
	        g_free (field);
		h_tuple_free (tuple);
		return g_string_free (string_out, FALSE);
	      }

	    /* Append metadata field */
	    for (n = 0; n < BMP_DATUM_NONE; n++)
	      {
		if (!g_ascii_strcasecmp (field, bmp_metadata_get_id_static(n)))
                  {
		    HTupleElem *elem = NULL;

		    elem = h_tuple_get (tuple, field);

		    if (elem)
		      {
			gchar *value = NULL; 

			if (elem->type == G_TYPE_INT)
			  {
			    value = g_strdup_printf ("%d", GPOINTER_TO_INT(elem->value));
			  }
			else if (elem->type == G_TYPE_STRING)
			  {
			    value = g_strdup (elem->value);
			  }
			
			if (value)
			  {
			    g_strstrip (value);
			    if (strlen(value)>0)
			      {
				hit = TRUE;
				g_string_append (string_out, value);
			      }
			    g_free (value);
			  }
		      }
		    break;
		  }
	      }

	    g_free (field);
	} else {
	    g_string_append_c (string_out, *format);
	}
	format++;
    }

    h_tuple_free (tuple);

    if (!hit)
      {
	gchar *filename, *basename;

	g_string_free (string_out, TRUE);

	if (!g_ascii_strncasecmp("file:",uri,5))
	  {
	    filename = g_filename_from_uri (uri, NULL, NULL);

	    if (filename)
	      {
		basename = g_path_get_basename (filename);
		g_free (filename);
		if (basename)
		  return basename;
	      }
	  }
	
	return g_strdup (uri);
      }
    else
	return g_string_free (string_out, FALSE);
}
