#include <string.h>
#include <assert.h>
#include <signal.h>
#include <getopt.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <unistd.h>

#include <glib.h>
#include <glib-object.h>

#include <config.h>

#include "main.h"
#include "play.h"
#include "configuration.h"

#include "bmpx/util.h"
#include "bmpx/uri.h"
#include "bmpx/metadata.h"

#include "marshalers.h"
#include "g_object_get.h"

#include <libhrel/types.h>
#include <libhrel/tuple.h>

#ifdef HAVE_XINE

#include <xine.h>

G_DEFINE_TYPE (BmpPlay, bmp_play, G_TYPE_OBJECT)

enum {
    BMP_PLAY_SIGNAL_EOS,
    BMP_PLAY_SIGNAL_ERROR,
    BMP_PLAY_SIGNAL_BUFFER,
    BMP_PLAY_SIGNAL_NO_SOUND_OUTPUT,
    BMP_PLAY_SIGNAL_TIME_TICK,

    BMP_PLAY_N_SIGNALS
};

static guint signals[BMP_PLAY_N_SIGNALS] = { 0, 0, 0, 0, 0 };

enum {
    BMP_PLAY_PROP_STREAM = 1,
    BMP_PLAY_PROP_PLAYSTATUS = 2,
    BMP_PLAY_N_PROPS
};

struct _BmpPlayPrivate {

    gboolean dispose_has_run;

    xine_t *xine;
    xine_stream_t *stream;
    xine_audio_port_t *audioport;
    xine_event_queue_t *evq;

    gchar *mrl;
    gint playstatus;
};

static GObject *bmp_play_constructor (GType type,
				      guint n_construct_properties,
				      GObjectConstructParam *
				      construct_properties);
static void send_stream_params (BmpPlay * self);

static gboolean
_idle_emit_eos (BmpPlay * self)
{
    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    g_signal_emit (G_OBJECT (self), signals[BMP_PLAY_SIGNAL_EOS], 0, NULL);

    return FALSE;
}

#if 0
static gboolean
_fadeout (BmpPlay * self)
{
    gint volume;

    volume = xine_get_param (self->private->stream, XINE_PARAM_AUDIO_VOLUME);

    volume -= 10;

    if (volume <= 0)
	{
	    xine_stop (self->private->stream);
	    xine_close (self->private->stream);
	    send_stream_params (self);
	    g_object_set (G_OBJECT (self), "playstatus", PLAYSTATUS_STOPPED,
			  NULL);
	    return FALSE;
	}
    else
	{
	    xine_set_param (self->private->stream, XINE_PARAM_AUDIO_VOLUME,
			    volume);
	    return TRUE;
	}
}
#endif

void
bmp_play_set_volume (BmpPlay * self, gint volume)
{
    g_return_if_fail (BMP_IS_PLAY (self));

    if (!self->private->stream)
	return;

    xine_set_param (self->private->stream, XINE_PARAM_AUDIO_VOLUME, volume);
}

void
bmp_play_set_eq (BmpPlay * self, BmpEqEnum eq, gint value)
{
    g_return_if_fail (BMP_IS_PLAY (self));

    if (!self->private->stream)
	return;

    xine_set_param (self->private->stream, eq + BMP_EQ_TO_ENGINE_OFFSET,
		    value);
}

#if 0
static void
_playobjects_destroy (BmpPlay * self)
{
    g_return_if_fail (BMP_IS_PLAY (self));
}
#endif

static void
event_listener_cb (gpointer user_data, const xine_event_t * event)
{
    switch (event->type)
	{
	case XINE_EVENT_UI_PLAYBACK_FINISHED:
		{
		    g_idle_add ((GSourceFunc) _idle_emit_eos, user_data);
		}
	    break;

	default:
	    break;
	}
}

static void
_playobjects_create (BmpPlay * self)
{
    gchar *filename;

    self->private->xine = xine_new ();

    filename = g_build_filename (g_get_home_dir (), ".xine", "config", NULL);
    xine_config_load (self->private->xine, filename);
    g_free (filename);
    xine_init (self->private->xine);

    self->private->audioport =
      xine_open_audio_driver (self->private->xine, NULL, NULL);
    self->private->stream =
      xine_stream_new (self->private->xine, self->private->audioport, NULL);
    self->private->evq = xine_event_new_queue (self->private->stream);
    xine_event_create_listener_thread (self->private->evq, event_listener_cb,
				       self);
}


gboolean
bmp_play_prepare (BmpPlay * self)
{
    return TRUE;
}

gboolean
bmp_play_stop_stream (BmpPlay * self)
{
    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    if (!self->private->stream)
	return TRUE;

    xine_stop (self->private->stream);
    xine_close (self->private->stream);
    send_stream_params (self);

    return TRUE;
}

#if 0
gboolean
bmp_play_stop_stream (BmpPlay * self)
{
    gint status;

    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    if (!self->private->stream)
	{
	    return TRUE;
	}

    status = xine_get_status (self->private->stream);

    if (status != XINE_STATUS_PLAY)
	return TRUE;

    g_timeout_add (75, _fadeout, self);

    return TRUE;
}
#endif

gboolean
bmp_play_ready (BmpPlay * self)
{
    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);
    return TRUE;
}

static void
send_stream_params (BmpPlay * self)
{
    gint status;
    uint32_t bitrate, samplerate;

    if (!self->private->stream)
	return;

    status = xine_get_status (self->private->stream);

    if (status != XINE_STATUS_STOP)
	{
	    bitrate =
	      xine_get_stream_info (self->private->stream,
				    XINE_STREAM_INFO_AUDIO_BITRATE);
	    bmp_system_control_bitrate (bmp_system_control, bitrate);

	    samplerate =
	      xine_get_stream_info (self->private->stream,
				    XINE_STREAM_INFO_AUDIO_SAMPLERATE);
	    bmp_system_control_samplerate (bmp_system_control, samplerate);
	}
    else
	{
	    bmp_system_control_bitrate (bmp_system_control, 0);
	    bmp_system_control_samplerate (bmp_system_control, 0);
	}
}

static gboolean
send_stream_pos (BmpPlay * self)
{
    gint stream_pos = bmp_play_get_time (self);
    uint32_t status;

    if (!self->private->stream)
	{

	    g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0,
			  NULL);
	    return FALSE;
	}

    status = xine_get_status (self->private->stream);

    if (status != XINE_STATUS_PLAY)
	{

	    g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0,
			  NULL);
	    g_object_set (G_OBJECT (self), "playstatus", PLAYSTATUS_STOPPED,
			  NULL);
#if 0
	    g_usleep (300000);
#endif
	    return FALSE;
	}

    if ((stream_pos >= 0) && (status == XINE_STATUS_PLAY))
	{

	    g_object_set (G_OBJECT (bmp_system_control), "stream_pos",
			  stream_pos, NULL);
	}
    else
	{
	    g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0,
			  NULL);
#if 0
	    g_usleep (300000);
#endif
	    return FALSE;
	}

    return TRUE;
}

gchar*
bmp_play_get_suffixes (BmpPlay *self)
{
  g_return_val_if_fail (BMP_IS_PLAY(self), NULL);

  return xine_get_file_extensions (self->private->xine);
}

gchar**
bmp_play_get_cd (BmpPlay * self, gint *n_mrls, gchar **disc_id)
{
    xine_stream_t *stream_aux;
    gchar **mrls;

    g_return_val_if_fail (BMP_IS_PLAY(self), NULL);

    mrls = 
	xine_get_autoplay_mrls (self->private->xine, "CD", n_mrls);
    stream_aux = 
	xine_stream_new (self->private->xine, NULL, NULL);
    xine_open (stream_aux, mrls[0]); 
    *disc_id = 
	g_strdup(xine_get_meta_info (stream_aux, XINE_META_INFO_CDINDEX_DISCID));
    xine_close (stream_aux);
    xine_dispose (stream_aux);

    return mrls;
}

gboolean
bmp_play_play_stream (BmpPlay * self)
{
    gchar *mrl_local;

    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    if (xine_get_param (self->private->stream, XINE_PARAM_SPEED) ==
	XINE_SPEED_PAUSE)
	{
	    xine_set_param (self->private->stream, XINE_PARAM_SPEED,
			    XINE_SPEED_NORMAL);
	    return TRUE;
	}

    xine_stop (self->private->stream);
    xine_close (self->private->stream);

    if (!g_ascii_strncasecmp (self->private->mrl, "cdda", 4))
      {
	mrl_local =
	    get_cdda_mrl_from_uri (self->private->mrl);
      }
    else
      {
	mrl_local =
	    g_strdup (self->private->mrl);
      }

    if (!xine_open (self->private->stream, mrl_local))
	{
	    g_message ("%s: %d", G_STRFUNC,
		       xine_get_error (self->private->stream));
	    bmp_system_message_send (bmp_system_control,
				     BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
	    g_free (mrl_local);
	    return TRUE;
	}

    if (!xine_play (self->private->stream, 0, 0))
	{
	    g_message ("%s: %d", G_STRFUNC,
		       xine_get_error (self->private->stream));
	    bmp_system_message_send (bmp_system_control,
				     BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
	    g_free (mrl_local);
	    return TRUE;
	}

    bmp_play_set_volume (self,
			 g_object_get_int (bmp_system_control, "volume"));

    send_stream_params (self);

    g_timeout_add_full (G_PRIORITY_DEFAULT, 250,
			(GSourceFunc) send_stream_pos, self, NULL);

    g_free (mrl_local);
    return TRUE;
}

Track*
bmp_play_get_metadata_xine (BmpPlay * self, const gchar *_uri) 
{
    GURI *g_uri;
    Track *track;
    xine_stream_t *stream_aux;
    gint stream_pos, stime, length;
    gchar *uri;
    const gchar* aux;

    g_return_val_if_fail (BMP_IS_PLAY(self), NULL);

    track = g_new0 (Track,1);

    track->mtime    = 0; 
    track->location = g_strdup (_uri);

    g_uri = bmp_uri_new (_uri);
    bmp_uri_unescape (g_uri);

    if (!strcmp("cdda",g_uri->scheme))
	uri = get_cdda_mrl_from_uri (_uri);
    else
	uri = g_strdup (_uri);

    stream_aux = xine_stream_new (self->private->xine, NULL, NULL);

    if (!xine_open (stream_aux, uri))
      {
	    g_message ("%s: could not open stream '%s'", G_STRLOC, uri);
	    g_free (uri);
	    return NULL;
      }

    xine_get_pos_length (stream_aux, &stream_pos, &stime, &length);
    aux = xine_get_meta_info (stream_aux, XINE_META_INFO_TRACK_NUMBER); 

    track->track      = aux ? strtol (aux, NULL, 10) : 0; 
    track->date       = 0;
    track->artist     = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_ARTIST)  ? xine_get_meta_info  (stream_aux, XINE_META_INFO_ARTIST)  : "(Unknown Artist)");
    track->album      = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_ALBUM)   ? xine_get_meta_info  (stream_aux, XINE_META_INFO_ALBUM)   : "(Unknown Album)");
    track->title      = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_TITLE)   ? xine_get_meta_info  (stream_aux, XINE_META_INFO_TITLE)   : "(Unknown Title)");
    track->genre      = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_GENRE)   ? xine_get_meta_info  (stream_aux, XINE_META_INFO_GENRE)   : "(Unknown Genre)");
    track->comment    = g_strdup (xine_get_meta_info (stream_aux, XINE_META_INFO_COMMENT) ? xine_get_meta_info  (stream_aux, XINE_META_INFO_COMMENT) : "(No Comment)");
    track->time       = length/1000; 
    track->rating     = 0;
    track->count      = 0;
    track->bitrate    = xine_get_stream_info (stream_aux, XINE_STREAM_INFO_AUDIO_BITRATE); 
    track->samplerate = xine_get_stream_info (stream_aux, XINE_STREAM_INFO_AUDIO_SAMPLERATE); 

    xine_close (stream_aux);
    xine_dispose (stream_aux);

    return track;
}

gboolean
bmp_play_pause_stream (BmpPlay * self)
{
    gint speed;

    speed = xine_get_param (self->private->stream, XINE_PARAM_SPEED);

    switch (speed)
	{

	case XINE_SPEED_PAUSE:
		{

		    g_object_set (self, "playstatus", PLAYSTATUS_PLAYING, NULL);
		}
	    break;

	case XINE_SPEED_NORMAL:
		{

		    xine_set_param (self->private->stream, XINE_PARAM_SPEED,
				    XINE_SPEED_PAUSE);
		}
	    break;
	}

    return TRUE;
}

gboolean
bmp_play_seek_stream (BmpPlay * self, gint offset)
{
    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    if (!self->private->stream)
	return FALSE;

    return !xine_play (self->private->stream, 0, offset) ? TRUE : FALSE;
}

gint
bmp_play_get_status (BmpPlay * self)
{
    gint status;

    if (!self->private->stream)
	return FALSE;

    status = xine_get_status (self->private->stream);

    if (status == XINE_STATUS_STOP)
	return PLAYSTATUS_STOPPED;
    if (status == XINE_STATUS_PLAY)
	return PLAYSTATUS_PLAYING;

    return status;
}

gint
bmp_play_get_time (BmpPlay * self)
{
    gint stream_pos, length, stime;

    if (!self->private->stream)
	return 0;

    xine_get_pos_length (self->private->stream, &stream_pos, &stime, &length);
    return stime / 1000;
}

gint
bmp_play_get_length (BmpPlay * self)
{
    gint stream_pos, length, stime;

    if (!self->private->stream)
	return 0;

    xine_get_pos_length (self->private->stream, &stream_pos, &stime, &length);

    return length / 1000;
}

static gboolean
playstatus_notify_cb (BmpPlay * self)
{
    gint status;

    g_return_val_if_fail (BMP_IS_PLAY (self), FALSE);

    g_object_get (self, "playstatus", &status, NULL);

    if (status & PLAYSTATUS_SEEKING)
	{
	    bmp_play_pause_stream (self);
	    return TRUE;
	}

    if (status & PLAYSTATUS_STOPPED)
	{
	    bmp_play_stop_stream (self);
	    g_object_set (G_OBJECT (bmp_system_control), "stream_pos", (gint) 0,
			  NULL);
	    return TRUE;
	}

    if (status & PLAYSTATUS_PAUSED)
	{
	    bmp_play_pause_stream (self);
	    return TRUE;
	}

    if (status & PLAYSTATUS_PLAYING)
	{
	    bmp_play_play_stream (self);
	    return TRUE;
	}

    return FALSE;
}

static void
bmp_play_set_property (GObject * object,
		       guint property_id,
		       const GValue * value, GParamSpec * pspec)
{
    BmpPlay *self = (BmpPlay *) object;

    switch (property_id)
	{
	case BMP_PLAY_PROP_PLAYSTATUS:
		{
		    gint status = g_value_get_int (value);

		    self->private->playstatus = status;
		}
	    break;

	case BMP_PLAY_PROP_STREAM:
		{
		    gchar *mrl;

		    if (self->private->mrl)
			g_free (self->private->mrl);

		    mrl = g_value_dup_string (value);

		    self->private->mrl = mrl;
		}
	    break;

	default:
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_play_get_property (GObject * object,
		       guint property_id, GValue * value, GParamSpec * pspec)
{
    BmpPlay *self = (BmpPlay *) object;

    switch (property_id)
	{
	case BMP_PLAY_PROP_STREAM:
		{
		    g_value_set_string (value, self->private->mrl);
		}
	    break;

	case BMP_PLAY_PROP_PLAYSTATUS:
		{
		    g_value_set_int (value, self->private->playstatus);
		}
	    break;

	default:
	    /*
	     * We don't have any other property...
	     */
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_play_dispose (GObject * obj)
{
    BmpPlay *self = (BmpPlay *) obj;

    if (self->private->dispose_has_run)
	return;

    self->private->dispose_has_run = TRUE;

    g_message ("%s\t\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_play_finalize (GObject * obj)
{
    BmpPlay *self = (BmpPlay *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free (self->private);
}

static void
bmp_play_class_init (BmpPlayClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    GParamSpec *pspec;

    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->dispose = bmp_play_dispose;
    gobject_class->finalize = bmp_play_finalize;
    gobject_class->constructor = bmp_play_constructor;

    pspec = g_param_spec_string ("mrl", "mrl", "mrl", "", G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, BMP_PLAY_PROP_STREAM,
				     pspec);

    pspec =
      g_param_spec_int ("playstatus",
			"playstatus",
			"player playing status",
			0, 128, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, BMP_PLAY_PROP_PLAYSTATUS,
				     pspec);

    signals[BMP_PLAY_SIGNAL_TIME_TICK] =
      g_signal_new ("time_tick",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, time_tick),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);


    signals[BMP_PLAY_SIGNAL_BUFFER] =
      g_signal_new ("buffer",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, buffer),
		    NULL, NULL, g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1, G_TYPE_POINTER);

    signals[BMP_PLAY_SIGNAL_EOS] =
      g_signal_new ("eos",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, eos),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
		    0);

    signals[BMP_PLAY_SIGNAL_ERROR] =
      g_signal_new ("error",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, error),
		    NULL, NULL, g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1, G_TYPE_POINTER);

    signals[BMP_PLAY_SIGNAL_NO_SOUND_OUTPUT] =
      g_signal_new ("no_sound",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, no_sound),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);
}

BmpPlay *
bmp_play_new (void)
{
    BmpPlay *self;

    self = g_object_new (BMP_TYPE_PLAY, NULL);
    return self;
}

static GObject *
bmp_play_constructor (GType type,
		      guint n_construct_properties,
		      GObjectConstructParam * construct_properties)
{
    GObject *obj;

	{
	    /*
	     * Invoke parent constructor.
	     */
	    BmpPlayClass *klass;
	    GObjectClass *parent_class;
	    klass = PLAY_CLASS (g_type_class_peek (BMP_TYPE_PLAY));
	    parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	    obj =
	      parent_class->constructor (type,
					 n_construct_properties,
					 construct_properties);
	}

    return obj;
}

static void
bmp_play_init (BmpPlay * self)
{
    self->private = g_new (BmpPlayPrivate, 1);

    self->private->dispose_has_run = FALSE;
    self->private->mrl = NULL;
    self->private->stream = NULL;
    self->private->audioport = NULL;
    self->private->xine = NULL;
    self->private->playstatus = PLAYSTATUS_STOPPED;
    self->private->evq = NULL;

    _playobjects_create (self);

    g_signal_connect (G_OBJECT (self), "notify::playstatus",
		      G_CALLBACK (playstatus_notify_cb), NULL);

    g_message ("%s\t\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
}

#endif

#ifdef HAVE_GST

#include <gst/gst.h>
#include <gst/gstelement.h>

G_DEFINE_TYPE (BmpPlay, bmp_play, G_TYPE_OBJECT)
	enum
{
    BMP_PLAY_SIGNAL_EOS,
    BMP_PLAY_SIGNAL_ERROR,
    BMP_PLAY_SIGNAL_BUFFER,
    BMP_PLAY_SIGNAL_NO_SOUND_OUTPUT,

    BMP_PLAY_N_SIGNALS
};

static guint signals[BMP_PLAY_N_SIGNALS] = { 0, 0, 0, 0 };

enum
{
    BMP_PLAY_PROP_STREAM = 1,
    BMP_PLAY_PROP_PLAYSTATUS = 2,
    BMP_PLAY_N_PROPS
};

struct _BmpPlayPrivate
{

    gboolean dispose_has_run;

    GstElement *playbin;
    GstElement *outsink;

    gchar *mrl;
    gint playstatus;
    gdouble volume;
};

static GObject *bmp_play_constructor (GType type,
				      guint n_construct_properties,
				      GObjectConstructParam *
				      construct_properties);
static GstBusSyncReply
bmp_play_bus_watch (GstBus * bus, GstMessage * message, BmpPlay * self);

#if 0
void
dn_caps_cb (GstElement * identity,
	    GObject * obj, GParamSpec * pspec, gpointer data)
{
    const GstCaps *caps;
    GstStructure *structure;
    gchar *s;
    gint chans = 0, rate = 0;

    g_return_if_fail (G_IS_OBJECT (obj));

    if (!GST_IS_PAD (obj) || !GST_PAD_IS_SINK (GST_PAD (obj)))
	return;

    if (!gst_pad_is_negotiated (GST_PAD (obj)))
	return;

    caps = gst_pad_get_negotiated_caps (GST_PAD (obj));

    s = gst_caps_to_string (caps);

    g_print ("Caps: %s\n", s);
    g_free (s);

    structure = gst_caps_get_structure (caps, 0);

    if (gst_structure_get_int (structure, "channels", &chans)
	&& gst_structure_get_int (structure, "rate", &rate))
	{
	    g_print ("Format: %s\nChannels: %d\nRate: %d\n",
		     gst_structure_get_name (structure), chans, rate);
	}
}
#endif

void
bmp_play_set_volume (BmpPlay * play, gint volume)
{
    gdouble _volume;

    g_return_if_fail (BMP_IS_PLAY (play));

    _volume = volume / 100.;

    g_object_set (play->private->playbin, "volume", _volume, NULL);
}


static void
_playobjects_destroy (BmpPlay * play)
{
    g_return_if_fail (BMP_IS_PLAY (play));
}

static void
_playobjects_create (BmpPlay * play)
{
    play->private->playbin = gst_element_factory_make ("playbin", "playbin");
    play->private->outsink =
      gst_element_factory_make ("alsasink", "audio-sink");

    g_object_set (G_OBJECT (play->private->outsink), "buffer-time",
		  (gint64) 50000, NULL);
    g_object_set (G_OBJECT (play->private->outsink), "latency-time",
		  (gint64) 500, NULL);

    g_object_set (G_OBJECT (play->private->playbin), "audio-sink",
		  play->private->outsink, NULL);
    g_object_set (G_OBJECT (play->private->playbin), "play-timeout",
		  (guint64) 3000000000, NULL);

    gst_bus_set_sync_handler (
          gst_element_get_bus (GST_ELEMENT (play->private->playbin)),
			 bmp_play_bus_watch, play);
}


#if 0
static gboolean
bmp_play_eos_cb (BmpPlay * play, GstElement * element)
{
    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    g_signal_emit (G_OBJECT (play), signals[BMP_PLAY_SIGNAL_EOS],
		   g_quark_from_string (""), NULL);

    return FALSE;
}
#endif

static void
bmp_play_error_cb (BmpPlay * play,
		   GstElement * arg1,
		   GError * arg2, gchar * arg3, GstElement * element)
{
    gchar *detail;

    if (!arg3)
	detail = g_strdup ("");
    else
	detail = g_strdup (arg3);

    g_signal_emit (G_OBJECT (play),
		   signals[BMP_PLAY_SIGNAL_ERROR],
		   g_quark_from_string (detail), arg2, NULL);

    g_idle_add ((GSourceFunc) _playobjects_destroy, play);
}

gboolean
bmp_play_prepare (BmpPlay * play)
{
    g_object_set (play->private->playbin, "uri", play->private->mrl, 0);
    return TRUE;
}

gboolean
bmp_play_stop_stream (BmpPlay * play)
{
    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    gst_element_set_state (play->private->playbin, GST_STATE_READY);
    g_usleep (50000);

    return TRUE;
}

gboolean
bmp_play_ready (BmpPlay * play)
{
    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    gst_element_set_state (play->private->playbin, GST_STATE_READY);

    return TRUE;
}

gboolean
bmp_play_play_stream (BmpPlay * play)
{

    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    bmp_play_stop_stream (play);
    if (gst_element_set_state (play->private->playbin, GST_STATE_PLAYING)
	== GST_STATE_CHANGE_FAILURE)
	{
	    gst_element_set_state (play->private->playbin, GST_STATE_READY);
	    return FALSE;
	}

    return TRUE;
}

/* Welcome to 'Reading metadata with GST' :P */

static GstElement *playbin_tags = NULL;

void
tag_foreach (const GstTagList * list, const gchar * tag,
	     GHashTable * metadata)
{
    GValue value = { 0, };
    gchar *aux;

    gst_tag_list_copy_value (&value, list, tag);

    if (G_VALUE_HOLDS (&value, G_TYPE_STRING))
	{
	    g_hash_table_insert (metadata, tag, g_value_dup_string (&value));
	    return;
	}

    if (G_VALUE_HOLDS (&value, G_TYPE_INT))
	{
	    aux = g_strdup_printf ("%d", g_value_get_int (&value));
	    g_hash_table_insert (metadata, tag, aux);
	    return;
	}

    if (G_VALUE_HOLDS (&value, G_TYPE_UINT))
	{
	    aux = g_strdup_printf ("%u", g_value_get_int (&value));
	    g_hash_table_insert (metadata, tag, aux);
	    return;
	}

}

static GstBusSyncReply
bus_watch_tags (GstBus * bus, GstMessage * message, GHashTable * metadata)
{
    GstState old_state, new_state;
    GstTagList *tag_list;


    switch (GST_MESSAGE_TYPE (message))
	{
	case GST_MESSAGE_TAG:
		{
		    gst_message_parse_tag (message, &tag_list);
		    gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach,
					  metadata);
		}
	    break;

	case GST_MESSAGE_ERROR:
	case GST_MESSAGE_WARNING:
		{
		    GError *err;
		    gchar *debug;

		    gst_message_parse_error (message, &err, &debug);
		    g_print ("%s ERROR: %s\n", G_STRFUNC, err->message);
		    g_error_free (err);
		    g_free (debug);
		    gst_element_set_state (playbin_tags, GST_STATE_READY);
		}
	    break;

	default:
	    /* unhandled message */
	    break;
	}

    /* remove message from the queue */
    return GST_BUS_DROP;
}

GHashTable *
bmp_play_get_metadata_gst (BmpPlay * self, const gchar * filename)
{
    GHashTable *metadata;
    gchar *mrl;

    playbin_tags = gst_element_factory_make ("playbin", "playbin");
    metadata = g_hash_table_new (g_str_hash, g_str_equal);

    gst_bus_set_sync_handler (gst_element_get_bus
			      (GST_ELEMENT (playbin_tags)),
			      (GstBusSyncHandler) bus_watch_tags, metadata);


    mrl = g_strconcat ("file://", filename, NULL);

    g_object_set (G_OBJECT (playbin_tags), "uri", mrl, NULL);
    gst_element_set_state (playbin_tags, GST_STATE_PAUSED);
    g_object_unref (G_OBJECT (playbin_tags));

    return metadata;
}

gboolean
bmp_play_pause_stream (BmpPlay * play)
{
    if (GST_STATE (play->private->playbin) == GST_STATE_PLAYING)
	return gst_element_set_state (play->private->playbin,
				      GST_STATE_PAUSED);
    else if (GST_STATE (play->private->playbin) == GST_STATE_PAUSED)
	return gst_element_set_state (play->private->playbin,
				      GST_STATE_PLAYING);
}

gboolean
bmp_play_seek_stream (BmpPlay * play, gint offset)
{
    GstEvent *event;
    GstFormat fmt = GST_FORMAT_TIME;
    guint64 s_offset;

    s_offset = offset * GST_SECOND;

    return gst_element_seek (play->private->playbin,
			     1.0,
			     fmt,
			     GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE,
			     GST_SEEK_TYPE_SET,
			     s_offset, GST_SEEK_TYPE_SET, s_offset);
}

GstState
bmp_play_get_element_state (BmpPlay * play)
{
    return GST_STATE (play->private->playbin);
}

gint
bmp_play_get_time (BmpPlay * play)
{
    GstQuery *query;
    GstFormat fmt = GST_FORMAT_TIME;
    gint64 time_in_nanoseconds;
    gint time_in_seconds;

    if (!(GST_STATE (play->private->playbin) == GST_STATE_PLAYING))
	return 0;

    query = gst_query_new_position (fmt);
    gst_element_query (play->private->playbin, query);
    gst_query_parse_position (query, &fmt, &time_in_nanoseconds);

    time_in_seconds = time_in_nanoseconds / 1000000000;

    return time_in_seconds;
}

gint
bmp_play_get_length (BmpPlay * play)
{
    GstQuery *query;
    GstFormat fmt = GST_FORMAT_TIME;
    gint64 time_in_nanoseconds, length_in_nanoseconds;
    gint time_in_seconds;

    if (!(GST_STATE (play->private->playbin) == GST_STATE_PLAYING))
	return 0;

    query = gst_query_new_position (fmt);
    gst_element_query (play->private->playbin, query);
    gst_query_parse_position (query, &fmt, 
			      &length_in_nanoseconds);

    length_in_nanoseconds = length_in_nanoseconds / 1000000000;

    return length_in_nanoseconds;
}

gint
bmp_play_get_status (BmpPlay * self)
{
    gint status;

    status = GST_STATE (self->private->playbin);

    if (status == GST_STATE_READY)
	return PLAYSTATUS_STOPPED;
    if (status == GST_STATE_PLAYING)
	return PLAYSTATUS_PLAYING;

    return status;
}

static gboolean
playstatus_notify_cb (BmpPlay * play)
{
    gint status;

    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    g_object_get (play, "playstatus", &status, NULL);

    if (status & PLAYSTATUS_SEEKING)
	{
	    bmp_play_pause_stream (play);
	    return TRUE;
	}

    if (status & PLAYSTATUS_STOPPED)
	{
	    bmp_play_stop_stream (play);
	    return TRUE;
	}

    if (status & PLAYSTATUS_PAUSED)
	{
	    bmp_play_pause_stream (play);
	    return TRUE;
	}

    if (status & PLAYSTATUS_PLAYING)
	{
	    bmp_play_play_stream (play);
	    return TRUE;
	}

    return FALSE;
}

static void
bmp_play_set_property (GObject * object,
		       guint property_id,
		       const GValue * value, GParamSpec * pspec)
{
    BmpPlay *play = (BmpPlay *) object;

    switch (property_id)
	{
	case BMP_PLAY_PROP_PLAYSTATUS:
		{
		    gint status = g_value_get_int (value);

		    play->private->playstatus = status;
		}
	    break;

	case BMP_PLAY_PROP_STREAM:
		{
		    gchar *stream_uri;

		    if (play->private->mrl)
			g_free (play->private->mrl);

		    stream_uri = g_value_dup_string (value);

		    play->private->mrl = g_strdup (stream_uri);

		    g_object_set (play->private->playbin, "uri",
				  play->private->mrl, 0);

		}
	    break;

	default:
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_play_get_property (GObject * object,
		       guint property_id, GValue * value, GParamSpec * pspec)
{
    BmpPlay *play = (BmpPlay *) object;

    switch (property_id)
	{
	case BMP_PLAY_PROP_STREAM:
		{
		    g_value_set_string (value, play->private->mrl);
		}
	    break;

	case BMP_PLAY_PROP_PLAYSTATUS:
		{
		    g_value_set_int (value, play->private->playstatus);
		}
	    break;

	default:
	    /*
	     * We don't have any other property...
	     */
	    g_assert (FALSE);
	    break;
	}
}

static void
bmp_play_dispose (GObject * obj)
{
    BmpPlay *play = (BmpPlay *) obj;

    if (play->private->dispose_has_run)
	return;

    play->private->dispose_has_run = TRUE;

    g_message ("%s\t\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(obj));
}

static void
bmp_play_finalize (GObject * obj)
{
    BmpPlay *play = (BmpPlay *) obj;

    /*
     * Here, complete object destruction.
     * You might not need to do much...
     */
    g_free (play->private);
}

static void
bmp_play_class_init (BmpPlayClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    GParamSpec *pspec;

    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->dispose = bmp_play_dispose;
    gobject_class->finalize = bmp_play_finalize;
    gobject_class->constructor = bmp_play_constructor;

    pspec = g_param_spec_string ("mrl", "mrl", "mrl", "", G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, BMP_PLAY_PROP_STREAM,
				     pspec);

    pspec =
      g_param_spec_int ("playstatus",
			"playstatus",
			"player playing status",
			0, 128, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, BMP_PLAY_PROP_PLAYSTATUS,
				     pspec);

    signals[BMP_PLAY_SIGNAL_BUFFER] =
      g_signal_new ("buffer",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, buffer),
		    NULL, NULL, g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1, G_TYPE_POINTER);

    signals[BMP_PLAY_SIGNAL_EOS] =
      g_signal_new ("eos",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, eos),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE,
		    0);

    signals[BMP_PLAY_SIGNAL_ERROR] =
      g_signal_new ("error",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, error),
		    NULL, NULL, g_cclosure_marshal_VOID__POINTER,
		    G_TYPE_NONE, 1, G_TYPE_POINTER);

    signals[BMP_PLAY_SIGNAL_NO_SOUND_OUTPUT] =
      g_signal_new ("no_sound",
		    G_OBJECT_CLASS_TYPE (gobject_class),
		    G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED,
		    G_STRUCT_OFFSET (BmpPlayClass, no_sound),
		    NULL, NULL, g_cclosure_marshal_VOID__VOID,
		    G_TYPE_NONE, 0);
}

BmpPlay *
bmp_play_new (void)
{
    BmpPlay *play;

    play = g_object_new (BMP_TYPE_PLAY, NULL);
    return play;
}

static GObject *
bmp_play_constructor (GType type,
		      guint n_construct_properties,
		      GObjectConstructParam * construct_properties)
{
    GObject *obj;

	{
	    /*
	     * Invoke parent constructor.
	     */
	    BmpPlayClass *klass;
	    GObjectClass *parent_class;
	    klass = PLAY_CLASS (g_type_class_peek (BMP_TYPE_PLAY));
	    parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
	    obj =
	      parent_class->constructor (type,
					 n_construct_properties,
					 construct_properties);
	}

    /*
     * do stuff.
     */
    return obj;
}

static gboolean
_idle_emit_eos (BmpPlay * play)
{
    g_return_val_if_fail (BMP_IS_PLAY (play), FALSE);

    g_signal_emit (G_OBJECT (play), signals[BMP_PLAY_SIGNAL_EOS], 0, NULL);

    return FALSE;
}

#define MAX_RESET_RETRIES 1

static GstBusSyncReply
bmp_play_bus_watch (GstBus * bus, GstMessage * message, BmpPlay * self)
{
    GstState old_state, new_state, pending_state;

    switch (GST_MESSAGE_TYPE (message))
	{
	case GST_MESSAGE_STATE_CHANGED:
		{
		    gst_message_parse_state_changed (message, &old_state,
						     &new_state, &pending_state);
		    g_message ("OLD_STATE: %d, NEW_STATE: %d", old_state,
			       new_state);
		}
	    break;

	case GST_MESSAGE_ERROR:
	case GST_MESSAGE_WARNING:
		{
		    GError *err;
		    gchar *debug;

		    gst_message_parse_error (message, &err, &debug);
		    g_print ("%s ERROR: %s\n", G_STRFUNC, err->message);
		    g_error_free (err);
		    g_free (debug);

		    g_message ("RESETTING Playbin");

#if 0
		    bmp_play_stop_stream (self);
		    if (reset_ctr < MAX_RESET_RETRIES)
			{
			    g_usleep (500000);
			    bmp_play_play_stream (self);
			    reset_ctr++;
			}
		    else
			{
#endif
			    gchar *uri;

			    g_object_get (G_OBJECT (self->private->playbin), "uri",
					  &uri, NULL);
			    gst_element_set_state (self->private->playbin,
						   GST_STATE_NULL);

			    _playobjects_create (self);

			    g_object_set (self->private->playbin, "uri", uri, NULL);

			    bmp_play_play_stream (self);
#if 0
			    reset_ctr = 0;
			}
#endif

		}
	    break;

	case GST_MESSAGE_EOS:
		{
		    /* end-of-stream */
		    g_message ("STOP stream");
		    g_idle_add ((GSourceFunc) _idle_emit_eos, self);
		}
	    break;

	default:
	    /* unhandled message */
	    break;
	}

    /* remove message from the queue */
    return GST_BUS_DROP;
}

static void
bmp_play_init (BmpPlay * play)
{
    play->private = g_new (BmpPlayPrivate, 1);

    play->private->dispose_has_run = FALSE;
    play->private->mrl = NULL;

    _playobjects_create (play);

    g_signal_connect (G_OBJECT (play), "notify::playstatus",
		      G_CALLBACK (playstatus_notify_cb), NULL);
}


#endif
