/*  BMP - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel Sikkes
 *
 *  ui.c/bmp_ui.h
 *  bmp_button.c/bmp_button.h
 *  bmp_button_toggle.c/bmp_button_toggle.h
 *  bmp_slider.c/bmp_slider.h
 *
 *  M.Derezynski
 *  Tim-Philipp Mueller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "main.h"
#include "error.h"
#include "resource_manager.h"

#include "ui.h"
#include "ui_playlist.h"
#include "ui_main.h"
#include "ui_util.h"
#include "ui_callbacks.h"
#include "ui_prefswin.h"

#include "bmp_button.h"
#include "bmp_button_toggle.h"
#include "bmp_slider.h"
#include "bmp_window.h"
#include "chroma_list.h"
#include "gtkstatusicon.h"

#include "bmpx/util.h"
#include "g_object_get.h"

G_DEFINE_TYPE (BmpUI, bmp_ui, G_TYPE_OBJECT)

/* GObject stuff */
enum {
    BMP_UI_SIGNAL_0,

    BMP_UI_N_SIGNALS
};

#if 0
static guint signals[BMP_UI_N_SIGNALS] = { 0 };
#endif

enum {
    BMP_PLAY_N_PROPS
};

struct _BmpUIPrivate {

    gboolean dispose_has_run;

    BmpWindowMain         *window_main;
    BmpWindowPlaylist     *window_playlist;
    BmpWindowPreferences  *window_preferences;

    GdkBitmap             *window_main_mask;

    GtkStatusIcon	  *status_icon;
    GdkPixbuf		  *status_icon_image_default;
    GdkPixbuf		  *status_icon_image_paused;
    GdkPixbuf		  *status_icon_image_playing;
    GdkPixbuf		  *status_icon_image_exit;
};

static gchar* bmp_skin_files[] = {
	"main.bmp",
	"shufrep.bmp",
	"cbuttons.bmp",
	"posbar.bmp",
	"volume.bmp",
	"titlebar.bmp",
	"balance.bmp",
	"pledit.bmp",
	"monoster.bmp",
	NULL,
	"playpaus.bmp"
};

static BmpSkinComponent bmp_skin_files_sequential[] = {
	SKIN_COMPONENT_MAIN,
	SKIN_COMPONENT_SHUFREP,
	SKIN_COMPONENT_CBUTTONS,
	SKIN_COMPONENT_TITLEBAR,
	SKIN_COMPONENT_PLEDIT,
	SKIN_COMPONENT_MONOSTER,
	SKIN_COMPONENT_PLAYPAUSE
};

static void
on_action_show_ui_toggled (GtkToggleAction *action, gpointer data);

static void
on_action_keep_above_toggled (GtkToggleAction *action, gpointer data);

static GtkToggleActionEntry bmp_toggle_actions_UI[] = {

  { BMP_TOGGLE_ACTION_SHOW_UI,
	NULL,
	N_("Show UI"),
	NULL,
	N_("Toggle BMPx UI Visiblity"),
	G_CALLBACK (on_action_show_ui_toggled),
	FALSE },

  { BMP_TOGGLE_ACTION_KEEP_ABOVE,
	NULL,
	N_("Keep BMP above other windows"),
	NULL,
	N_("Keep BMP above other windows"),
	G_CALLBACK (on_action_keep_above_toggled),
	FALSE }

};

/* 'extern' referenced */
GtkUIManager *bmp_ui_manager = NULL;
BmpChromaWidget *chroma_widgets[BMP_N_CHROMA_WIDGETS];
bmp_posbar_t *bmp_posbar;
bmp_volume_t *bmp_volume;
bmp_volume_t *bmp_balance;
bmp_skin_component_t *bmp_skin_component[N_SKIN_COMPONENTS];

/* Action groups */
GtkActionGroup *bmp_actions = NULL;
GtkActionGroup *bmp_toggle_actions = NULL;

static void
free_skin_components(BmpUI *self)
{
  gint n;

 /* Free memory */
  for (n = 0; n < N_SKIN_COMPONENTS; n++)
  {
     if (bmp_skin_component[n] != NULL)
     {
	g_object_unref (bmp_skin_component[n]->pixbuf);
	g_free (bmp_skin_component[n]);
	bmp_skin_component[n] = NULL;
     }
  }

  if (bmp_posbar)
	{
		g_object_unref (bmp_posbar->pixbuf);
		g_free (bmp_posbar);
		bmp_posbar = NULL;
	}

  if (bmp_volume)
	{
		g_object_unref (bmp_volume->pixbuf);
		g_free (bmp_volume);
		bmp_volume = NULL;
	}

  if (bmp_balance)
	{
		g_object_unref (bmp_balance->pixbuf);
		g_free (bmp_balance);
		bmp_balance = NULL;
	}
}

static void
load_skin_components(BmpUI *self, const gchar * archive_filename, gint type)
{
  GdkPixbuf *pixbuf;
  const gchar *archive_dirname;
  gchar *filename;
  gint n;

  /* Free memory */
  free_skin_components (self);
  bmp_cursors_free ();

  /* LOAD COMPONENTS */
  if (type == BMP_SKIN_ARCHIVE)
  	archive_dirname = archive_decompress (archive_filename);
  else
  	archive_dirname = archive_filename;

  bmp_cursors_set (archive_dirname);

  /* Buttons, Titlebars */
  for (n = 0; n < G_N_ELEMENTS(bmp_skin_files_sequential); n++)
    {
       filename = find_file_recursively(archive_dirname, bmp_skin_files[bmp_skin_files_sequential[n]], 1);
      if (!filename)
	{
        	g_message("%s not found", bmp_skin_files[bmp_skin_files_sequential[n]]);
	}
      else
        {
	      pixbuf = gdk_pixbuf_new_from_file(filename, NULL);

	      bmp_skin_component[bmp_skin_files_sequential[n]] = g_malloc0(sizeof(bmp_skin_component_t));
	      bmp_skin_component[bmp_skin_files_sequential[n]]->identifier = bmp_skin_files_sequential[n];
	      bmp_skin_component[bmp_skin_files_sequential[n]]->pixbuf = pixbuf;

	      g_free(filename);
        }
    }

  /* Numbers */
  bmp_skin_component[SKIN_COMPONENT_NUMBERS] = g_malloc0(sizeof(bmp_skin_component_t));
  filename = find_file_recursively(archive_dirname, "numbers.bmp", 1);
  if (!filename)
	filename = find_file_recursively(archive_dirname, "nums_ex.bmp", 1);

  if (!filename)
    {
	g_message("%s: Numbers not found", G_STRFUNC);
	g_free(bmp_skin_component[SKIN_COMPONENT_NUMBERS]);
    }
  else 
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_skin_component[SKIN_COMPONENT_NUMBERS]->identifier = SKIN_COMPONENT_NUMBERS;
	bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf = pixbuf;

	if (gdk_pixbuf_get_width(pixbuf) >= 108)
	   bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (12);
	else
	   bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data = GINT_TO_POINTER (10);

	g_free(filename);
    } 

  /* Volume, Balance and Seek slider*/
  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_POSBAR], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_POSBAR]);
    }
  else
   {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_posbar = g_malloc0(sizeof(bmp_posbar_t));
	bmp_posbar->pixbuf = pixbuf;
	g_free(filename);
   }

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_VOLUME], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_VOLUME]);
    }
  else
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_volume = g_malloc0(sizeof(bmp_volume_t));
	bmp_volume->pixbuf = pixbuf;
	bmp_volume->height = gdk_pixbuf_get_height(pixbuf) - 422;

	if (gdk_pixbuf_get_height(pixbuf) > 422)
	    bmp_volume->has_button = TRUE;
	else
	    bmp_volume->has_button = FALSE;
	g_free(filename);
    }

  filename = find_file_recursively(archive_dirname, bmp_skin_files[SKIN_COMPONENT_BALANCE], 1);
  if (!filename)
    {
	g_message("%s not found", bmp_skin_files[SKIN_COMPONENT_BALANCE]);
    }
  else
    {
	pixbuf = gdk_pixbuf_new_from_file(filename, NULL);
	bmp_balance = g_malloc0(sizeof(bmp_volume_t));
	bmp_balance->pixbuf = pixbuf;
	bmp_balance->height = gdk_pixbuf_get_height(pixbuf) - 421;
	if (gdk_pixbuf_get_height(pixbuf) > 421)
	    bmp_balance->has_button = TRUE;
	else
	    bmp_balance->has_button = FALSE;
	g_free(filename);
    }

  bmp_colors_init(archive_dirname);

  /* FIXME: Move this into ui_main.c */
  self->private->window_main_mask = bmp_get_mask_window_main (archive_dirname);

  if (type == BMP_SKIN_ARCHIVE)
    {
  	del_directory(archive_dirname);
    } 
}

GdkBitmap*
bmp_ui_get_mask_window_main (BmpUI *self)
{
    g_return_val_if_fail (BMP_IS_UI (self), NULL);

    return self->private->window_main_mask;
}

void
bmp_ui_display_windows(BmpUI *self) {

    gint x,y,w,h;
    gboolean visible;

    x = bmp_configuration_key_fetch_int (bmp_configuration, "main-window", "pos_x");
    y = bmp_configuration_key_fetch_int (bmp_configuration, "main-window", "pos_y");
    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH/2;
	y -= BMP_WINDOW_MAIN_HEIGHT/2;

	gtk_window_move (GTK_WINDOW(self->private->window_main->window), x, y);
	BMP_WINDOW (self->private->window_main->window)->x = x;
	BMP_WINDOW (self->private->window_main->window)->y = y;
    }
    gtk_widget_show_all (BMP_WINDOW_MAIN(self->private->window_main)->window);

    x = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "pos_x");
    y = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "pos_y");

    if ((x > -1) && (y > -1))
    {
	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }
    else
    {
	x = gdk_screen_get_width (gdk_screen_get_default());
	y = gdk_screen_get_height (gdk_screen_get_default());

	x /= 2;
	y /= 2;

	x -= BMP_WINDOW_MAIN_WIDTH;
	y -= BMP_WINDOW_MAIN_HEIGHT;

	gtk_window_move (GTK_WINDOW(self->private->window_playlist->window), x, y);
	BMP_WINDOW (self->private->window_playlist->window)->x = x;
	BMP_WINDOW (self->private->window_playlist->window)->y = y;
    }

    w = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "width");
    h = bmp_configuration_key_fetch_int (bmp_configuration, "playlist-window", "height");

    w = (w < 0) ? 10*25 : w;
    h = (h < 0) ? 10*29 : h;

#if 0
    gtk_widget_set_size_request (GTK_WINDOW(self->private->window_playlist->window), w, h); 
#endif
    gtk_window_set_default_size (GTK_WINDOW(self->private->window_playlist->window), w, h); 

    visible = bmp_configuration_key_fetch_boolean (bmp_configuration,"playlist-window","visible");

    if (visible)
	bmp_system_control_show_pl_window (bmp_system_control, TRUE);
}

static gint
load_skin(BmpUI *self, const gchar * filename)
{
    gint error_status;

    if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
	error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    bmp_colors_free();

    if (g_file_test(filename, G_FILE_TEST_IS_DIR)) {
        load_skin_components(self, filename, BMP_SKIN_DIR);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    if (g_file_test(filename, G_FILE_TEST_IS_REGULAR)) {
        load_skin_components(self, filename, BMP_SKIN_ARCHIVE);
	error_status = BMP_ERROR_SUCCESS;
    }
    else
    {
        error_status = BMP_ERROR_FAILURE;
	return error_status;
    }

    if (self->private->window_main)
        bmp_window_main_configure (BMP_WINDOW_MAIN (self->private->window_main));

    if (self->private->window_playlist)
	bmp_window_playlist_configure (BMP_WINDOW_PLAYLIST (self->private->window_playlist));

    return error_status;
}

static void
on_bmp_skin_changed (GObject *configuration, gpointer data)
{
    if (load_skin (BMP_UI(data), bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin")) != BMP_ERROR_SUCCESS) 
      {
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "Unable to load skin '%s'", bmp_configuration_key_fetch (bmp_configuration, "bmp", "skin"));
      }
}

static void
font_dispose (BmpFontInformation *fi)
{
    pango_font_description_free (fi->font_description);
    g_free (fi);
}

static void
on_bmp_titlestring_format_changed (BmpUI *self)
{
    gchar *title;
    g_return_if_fail (BMP_IS_UI(self));

    title = bmp_playlist_get_current_title (bmp_playlist);
    bmp_ui_set_trayicon_text (self, title);
    g_free (title);
}
 
static void
on_bmp_font_changed (BmpUI *self)
{
    PangoFontDescription *font_desc;
    const gchar *font;

    font = bmp_configuration_key_fetch (bmp_configuration, "bmp", "font");

    if (bmp_fonts[BMP_FONT_NORMAL] != NULL)
	font_dispose (bmp_fonts[BMP_FONT_NORMAL]);
    bmp_fonts[BMP_FONT_NORMAL] =
	bmp_get_font_information (font, self->private->window_main->window);

    font_desc =
	pango_font_description_copy (bmp_fonts[BMP_FONT_NORMAL]->font_description);

    pango_font_description_set_size (font_desc,
	pango_font_description_get_size(font_desc)+ (1.2*PANGO_SCALE));

    if (bmp_fonts[BMP_FONT_LARGE] != NULL)
	font_dispose (bmp_fonts[BMP_FONT_LARGE]);
    bmp_fonts[BMP_FONT_LARGE]  =
	bmp_get_font_information_pfd (font_desc, self->private->window_main->window);

    pango_font_description_free (font_desc);
}

void
bmp_ui_set_trayicon_exit (BmpUI *self)
{
    g_return_if_fail (BMP_IS_UI(self));

#if 0
    gtk_status_icon_set_pixbuf (self->private->status_icon, 
				   self->private->status_icon_image_exit);
#endif
    gtk_status_icon_set_blinking (self->private->status_icon, TRUE);
}

static void
on_tray_icon_click (GtkStatusIcon *icon, gpointer data)
{
    BmpUI *ui = BMP_UI(data);
    GtkToggleAction *action;
    gboolean active;
    
    action = GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_SHOW_UI));
    active = gtk_toggle_action_get_active (action);

    if ((active) &&
	(gdk_window_get_state (GTK_WIDGET(ui->private->window_main->window)->window) & GDK_WINDOW_STATE_ICONIFIED))
      {
	gdk_window_show (GTK_WIDGET(ui->private->window_main->window)->window);
      }
    else
      {
	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(action), !active);
      } 
}

static void
on_tray_icon_scroll_up (GtkStatusIcon *icon, gpointer data)
{
    gint volume;

    volume = g_object_get_int (bmp_system_control, "volume");
    volume += 4;
    volume = (volume > 100) ? 100 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
}

static void
on_tray_icon_scroll_down (GtkStatusIcon *icon, gpointer data)
{
    gint volume;

    volume = g_object_get_int (bmp_system_control, "volume");
    volume -= 4;
    volume = (volume < 0) ? 0 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
}


static void
on_tray_icon_popup_menu (GtkStatusIcon *icon, guint arg1, guint arg2, gpointer data)
#if 0
on_tray_icon_click (GtkStatusIcon *icon, gpointer data)
#endif
{
    GtkWidget *menu;

    menu = ui_manager_get_popup (bmp_ui_manager, "/popup-tray/menu-tray");
    gtk_widget_realize (menu);
    gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, arg1, arg2);
}

static void
on_system_control_set_playstatus (BmpSystemControl *system_control, gint status, gpointer data) 
{
    g_return_if_fail (BMP_IS_UI(data));

    if ((status != PLAYSTATUS_PLAYING) && (status != PLAYSTATUS_PAUSED))
      {
	    gtk_status_icon_set_from_pixbuf (BMP_UI(data)->private->status_icon, 
					   BMP_UI(data)->private->status_icon_image_default);
      }
    else if (status == PLAYSTATUS_PAUSED)
      {
	    gtk_status_icon_set_from_pixbuf (BMP_UI(data)->private->status_icon, 
					   BMP_UI(data)->private->status_icon_image_paused);
      }
    else if (status == PLAYSTATUS_PLAYING)
      {
	    gtk_status_icon_set_from_pixbuf (BMP_UI(data)->private->status_icon, 
					   BMP_UI(data)->private->status_icon_image_playing);
      }
}

static void
on_action_keep_above_toggled (GtkToggleAction *action, gpointer data)
{
    gtk_window_set_keep_above (GTK_WINDOW(bmp_ui->private->window_main->window),
                               gtk_toggle_action_get_active (action));
    gtk_window_set_keep_above (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                               gtk_toggle_action_get_active (action));
}

static void
on_action_show_ui_toggled (GtkToggleAction *action, gpointer data)
{
    gboolean active;
    static gboolean playlist_visible = FALSE;

    active = gtk_toggle_action_get_active (action);

    if (active)
      {
	gtk_widget_show (bmp_ui->private->window_main->window);
	if (playlist_visible)
	    gtk_widget_show (bmp_ui->private->window_playlist->window);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             FALSE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             FALSE); 

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             FALSE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             FALSE); 
	bmp_window_playlist_resize_widgets (bmp_ui->private->window_playlist);
      }
    else
      {
	gtk_widget_hide (GTK_WIDGET(bmp_ui->private->window_main->window));

	playlist_visible = GTK_WIDGET_VISIBLE(bmp_ui->private->window_playlist->window);
	if (playlist_visible)
	    gtk_widget_hide (bmp_ui->private->window_playlist->window);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             TRUE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_main->window),
                                             TRUE); 

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             TRUE); 
	gtk_window_set_skip_pager_hint  (GTK_WINDOW(bmp_ui->private->window_playlist->window),
                                             TRUE); 
      }
}


static gint
ui_initialize(BmpUI *self)
{
  GError *error = NULL;
  const gchar *filename, *listname; 
  gchar *title, *trayicon_ui;
  gint rv, n;

  if ((self->private->window_playlist != NULL) && (self->private->window_main != NULL)) return BMP_ERROR_SUCCESS;

  filename =
	bmp_configuration_key_fetch(bmp_configuration, "bmp", "skin");

  for (n = 0; n < N_SKIN_COMPONENTS; bmp_skin_component[n++] = NULL);

  bmp_cursors_init();

  rv =
    load_skin (self, filename);

  if (rv != BMP_ERROR_SUCCESS) {
	filename =
	    DATA_DIR G_DIR_SEPARATOR_S BMP_SKIN_DIR_BASENAME G_DIR_SEPARATOR_S BMP_DEFAULT_SKIN;
	bmp_configuration_key_set_string (bmp_configuration, "bmp","skin", filename);
	rv =
	    load_skin (self, filename);
  }

  if (rv == BMP_ERROR_SUCCESS) {
    gint n;

    g_object_connect (G_OBJECT(bmp_configuration),
			"signal::changed::bmp-skin",
			G_CALLBACK(on_bmp_skin_changed),
			self,
			"swapped-signal::changed::bmp-font",
			G_CALLBACK(on_bmp_font_changed),
			self,
			"swapped-signal::changed::bmp-titlestring-format",
			G_CALLBACK(on_bmp_titlestring_format_changed),
			self,
			NULL);

    for (n = 0; n < BMP_N_FONTS; bmp_fonts[n++] = NULL);

    bmp_ui_manager = gtk_ui_manager_new();

    bmp_resource_manager_create_path (bmp_rm, NULL, "windows");
    bmp_resource_manager_create_path (bmp_rm, NULL, "tooltips");

    bmp_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP);
    bmp_toggle_actions = gtk_action_group_new(BMP_MAIN_ACTION_GROUP_TOGGLE);

    gtk_action_group_set_translate_func (bmp_actions,
		bmp_menu_translate,
		NULL,
		NULL);
    gtk_action_group_set_translate_func (bmp_toggle_actions,
		bmp_menu_translate,
		NULL,
		NULL);

    gtk_ui_manager_insert_action_group (bmp_ui_manager, bmp_actions, 0);
    gtk_ui_manager_insert_action_group (bmp_ui_manager, bmp_toggle_actions, 0);
    gtk_action_group_set_sensitive (bmp_actions, TRUE);
    gtk_action_group_set_sensitive (bmp_toggle_actions, TRUE);

    gtk_action_group_add_toggle_actions (bmp_toggle_actions, bmp_toggle_actions_UI, G_N_ELEMENTS (bmp_toggle_actions_UI), self);

    self->private->window_main = bmp_window_main_new();
    RM_ITEM_NEW("windows", "window-main", self->private->window_main, NULL);
    g_signal_emit_by_name (G_OBJECT(bmp_configuration), "changed::bmp-font");

    self->private->window_playlist = bmp_window_playlist_new();
    RM_ITEM_NEW("windows", "window-playlist", self->private->window_playlist, NULL);
    self->private->window_preferences = bmp_window_preferences_new();
    RM_ITEM_NEW("windows", "window-preferences", self->private->window_preferences, NULL);

    /* Set up configuration bindings */
    bmp_configuration_bind_toggle_action (bmp_configuration, 
					GTK_TOGGLE_ACTION(gtk_action_group_get_action (bmp_toggle_actions, BMP_TOGGLE_ACTION_PL)),
					"playlist-window",
					"visible");

    bmp_window_main_configure (self->private->window_main);
    bmp_window_playlist_configure (self->private->window_playlist);

    bmp_ui_display_windows (self);

    chroma_list_set_list (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), 
			  BMP_LIST(bmp_playlist_get_playlists(bmp_playlist)));

    listname = bmp_configuration_key_fetch(bmp_configuration,"bmp","current-playlist");
    chroma_list_select_name (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), listname);

    bmp_system_control_repeat_set (bmp_system_control, 
				   bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "repeat"), 
				   NULL);
    bmp_system_control_shuffle_set (bmp_system_control, 
				    bmp_configuration_key_fetch_boolean (bmp_configuration, "bmp", "shuffle"), 
				    NULL);

    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));
    chroma_list_list_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));

    /* Init Tray Icon */
    self->private->status_icon_image_default =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-default.png",
				  NULL);
    self->private->status_icon_image_paused =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-paused.png",
				  NULL);
    self->private->status_icon_image_playing =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-playing.png",
				  NULL);
    self->private->status_icon_image_exit =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-exit.png",
				  NULL);
    self->private->status_icon = 
	gtk_status_icon_new_from_pixbuf (self->private->status_icon_image_default);
    gtk_status_icon_set_visible (self->private->status_icon, TRUE);

    g_object_connect (G_OBJECT (self->private->status_icon),
		    "signal::click",
		    G_CALLBACK(on_tray_icon_click),
		    self,
		    "signal::popup-menu",
		    G_CALLBACK(on_tray_icon_popup_menu),
		    self,
		    "signal::scroll-up",
		    G_CALLBACK(on_tray_icon_scroll_up),
		    self,
		    "signal::scroll-down",
		    G_CALLBACK(on_tray_icon_scroll_down),
		    self,
		    NULL);

    trayicon_ui = g_build_filename (DATA_DIR, BMP_UI_BASENAME, "trayicon.ui", NULL);

    if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager, trayicon_ui, &error))
      {
		g_message ("building menus %s failed: %s", trayicon_ui, error->message);
		g_error_free (error);
      }
    g_free (trayicon_ui);

    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action 
						(bmp_toggle_actions, BMP_TOGGLE_ACTION_SHOW_UI)), TRUE);

    gtk_toggle_action_set_active (GTK_TOGGLE_ACTION(gtk_action_group_get_action 
						(bmp_toggle_actions, BMP_TOGGLE_ACTION_KEEP_ABOVE)),
						 bmp_configuration_key_fetch_boolean (bmp_configuration,"bmp","keep-above"));

    g_object_connect (G_OBJECT(bmp_system_control),
			"signal::set-playstatus",
			G_CALLBACK(on_system_control_set_playstatus),
			self,
			NULL);

    title = bmp_playlist_get_current_title (bmp_playlist);
    bmp_ui_set_trayicon_text (bmp_ui, title); 
    g_free (title);
  }

  // DEBUG
  g_message ("%s: screen resolution: %f", G_STRLOC, screen_get_resolution(gdk_screen_get_default()));

  return rv;
}

/* GObject stuff */
static void
bmp_ui_dispose (GObject *obj)
{
    BmpUI *self = BMP_UI (obj);

    if (self->private->dispose_has_run)
         return;

    self->private->dispose_has_run = TRUE;

    free_skin_components (self);

    if (self->private->window_main)
      {
	gtk_widget_destroy (BMP_WINDOW(self->private->window_main->window)->canvas);
	gtk_widget_destroy (BMP_WINDOW(self->private->window_playlist->window)->canvas);
	BMP_WINDOW(self->private->window_main->window)->canvas = NULL;
	BMP_WINDOW(self->private->window_playlist->window)->canvas = NULL;

	gtk_widget_destroy (self->private->window_main->window);
	gtk_widget_destroy (self->private->window_playlist->window);

	g_object_unref (self->private->window_main);
	g_object_unref (self->private->window_playlist);
	g_object_unref (self->private->window_preferences);
	g_object_unref (bmp_ui_manager);

	bmp_ui_manager = NULL;
	self->private->window_main = NULL;
	self->private->window_playlist = NULL;
	self->private->window_preferences = NULL;
      }

    g_message("%s\t\t\t[DISPOSED]", G_OBJECT_TYPE_NAME(self));
}

static void
bmp_ui_class_init (BmpUIClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
    gobject_class->finalize = bmp_play_finalize;
    gobject_class->constructor = bmp_play_constructor;
#endif

    gobject_class->dispose = bmp_ui_dispose;
}

static void
bmp_ui_init (BmpUI *self)
{
    self->private = g_new0 (BmpUIPrivate, 1);

    self->private->window_main = NULL;
    self->private->window_playlist = NULL;
    self->private->window_preferences = NULL;

    self->private->window_main_mask = NULL;
}

BmpUI*
bmp_ui_new (void)
{
    BmpUI *self;

    self = g_object_new (BMP_TYPE_UI, NULL);
    return self;
}

gint
bmp_ui_finalize_init (BmpUI *self)
{
    gint rv;

    rv = ui_initialize(self);

    if (rv == BMP_ERROR_SUCCESS)
	    g_message("%s\t\t\t[INITIALIZED]", G_OBJECT_TYPE_NAME(self));
    else
	    g_message("%s:\t\t\tERROR: Not initalized: Skin couldn't be loaded", __FILE__);

    return rv;
}

void
bmp_ui_set_trayicon_text (BmpUI *self, const gchar *text)
{
    const gchar *stock_id;
    gint playstatus;

    g_return_if_fail (BMP_IS_UI(self));

    playstatus = g_object_get_int (bmp_play, "playstatus");

    switch (playstatus)
     {
	case PLAYSTATUS_STOPPED: stock_id = GTK_STOCK_MEDIA_STOP; break;
	case PLAYSTATUS_PAUSED:  stock_id = GTK_STOCK_MEDIA_PAUSE; break;
	case PLAYSTATUS_PLAYING: stock_id = GTK_STOCK_MEDIA_PLAY; break;
	default: stock_id = GTK_STOCK_MEDIA_STOP; break;
     }

    gtk_status_icon_set_tooltip (self->private->status_icon, text, stock_id);    
}
